# -*- coding: utf-8 -*-
"""
Created on Thu Jan 18 12:25:16 2024

@author: dvezinet
"""


import os
import getpass
import warnings
import datetime as dtm


import numpy as np
import matplotlib.colors as mcolors
import datastock as ds


# #################################################################
# #################################################################
#          Default values
# #################################################################


_COLOR = 'k'
_NAME = 'rays'


# #################################################################
# #################################################################
#          Main
# #################################################################


def main(
    # ---------------
    # input from tofu
    coll=None,
    key=None,
    key_cam=None,
    # input from arrays
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # input from file
    pfe_in=None,
    # ---------------
    # options
    outline_only=None,
    include_centroid=None,
    factor=None,
    color=None,
    # ---------------
    # saving
    pfe_save=None,
    overwrite=None,
):
    """ Export a set of LOS to a stp file (for CAD compatibility)

    The LOS can be provided either as:
        - (coll, key): if you're using tofu
        - pfe_in: a path-filename-extension to a valid csv or dat file

    In the second case, the file is assumed to hold a (nlos*npts, 3) array
        with a heaer speciying nlos and npts

    Parameters
    ----------
    coll : tf.data.Collection, optional
        DESCRIPTION. The default is None.
    key : str, optional
        DESCRIPTION. The default is None.
    pfe_in : str, optional
        DESCRIPTION. The default is None.
    outline_only: bool, optional
        Whether to keep only the outline of a 2d pixel map (default: True)
    factor : float, optional
        scaling factor on coordinates (default: 1)
    color : str / tuple, optional
        DESCRIPTION. The default is None.
    pfe_save : str, optional
        DESCRIPTION. The default is None.
    overwrite : bool, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """

    # ----------------
    # check inputs
    # --------------

    (
        key, key_cam,
        ptsx, ptsy, ptsz,
        pfe_in,
        outline_only,
        include_centroid,
        factor, color,
        iso,
        pfe_save, overwrite,
    ) = _check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        pfe_in=pfe_in,
        # options
        outline_only=outline_only,
        include_centroid=include_centroid,
        factor=factor,
        color=color,
        # saving
        pfe_save=pfe_save,
        overwrite=overwrite,
    )

    fname = os.path.split(pfe_save)[-1][:-4]

    # -------------
    # extract and pre-format data
    # -------------

    dptsx, dptsy, dptsz = _extract(
        coll=coll,
        key=key,
        key_cam=key_cam,
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        outline_only=outline_only,
        include_centroid=include_centroid,
        pfe_in=pfe_in,
        fname=fname,
    )

    # scaling factor
    for k0 in dptsx.keys():
        dptsx[k0] = factor * dptsx[k0]
        dptsy[k0] = factor * dptsy[k0]
        dptsz[k0] = factor * dptsz[k0]

    # ---------------
    # get color dict
    # ---------------

    dcolor = _get_dcolor(dptsx=dptsx, color=color)

    # ----------------
    # get file content
    # ----------------

    # HEADER
    msg_header = _get_header(
        fname=fname,
        iso=iso,
    )

    # DATA
    msg_data = _get_data(
        dptsx=dptsx,
        dptsy=dptsy,
        dptsz=dptsz,
        fname=fname,
        # options
        dcolor=dcolor,
        # norm
        iso=iso,
    )

    # -------------
    # save to stp
    # -------------

    _save(
        msg=msg_header + "\n" + msg_data,
        pfe_save=pfe_save,
        overwrite=overwrite,
    )

    return


# #################################################################
# #################################################################
#          check
# #################################################################


def _check(
    coll=None,
    key=None,
    key_cam=None,
    ptsx=None,
    ptsy=None,
    ptsz=None,
    pfe_in=None,
    # options
    outline_only=None,
    include_centroid=None,
    factor=None,
    color=None,
    # saving
    pfe_save=None,
    overwrite=None,
):

    # --------------
    # coll vs pfe_in
    # -------------

    lc = [coll is not None, ptsx is not None, pfe_in is not None]
    if np.sum(lc) != 1:
        msg = (
            "Please provide eiter a (Collection, key) pair xor array xor pfe_in!\n"
            f"\t- coll is None: {coll is None}\n"
            f"\t- (ptsx, ptsy, ptsz) is None: {ptsx is None}\n"
            f"\t- pfe_in is None: {pfe_in is None}\n"
        )
        raise Exception(msg)

    # ---------------
    # coll
    # ---------------

    if lc[0]:


        # ------------
        # coll

        if not issubclass(coll.__class__, ds.DataStock):
            msg = (
                "Arg coll must be a subclass of datastock.Datastock!\n"
                f"\t- type(coll) = {type(coll)}"
            )
            raise Exception(msg)

        # --------------
        # key

        lok_rays = list(coll.dobj.get('rays', {}).keys())
        lok_diag = list(coll.dobj.get('diagnostic', {}).keys())
        key = ds._generic_check._check_var(
            key, 'key',
            types=str,
            allowed=lok_rays + lok_diag,
        )

        if key in lok_diag:
            if isinstance(key_cam, str):
                key_cam = [key_cam]

            lok = coll.dobj['diagnostic'][key]['camera']
            key_cam = ds._generic_check._check_var_iter(
                key_cam, 'key_cam',
                types=(list, tuple),
                types_iter=str,
                allowed=lok,
                default=lok,
            )

        else:
            key_cam = None

    # ---------------
    # array
    # ---------------

    elif lc[1]:

        c0 = all([
            isinstance(pp, np.ndarray)
            and pp.ndim >= 2
            and pp.shape[0] >= 2
            for pp in [ptsx, ptsy, ptsz]
        ])
        if not c0:
            msg = (
                "Args (ptsx, ptsy, ptsz) must be np.ndarrays of shape (npts>=2, nlos)\n"
                f"\t- ptsx: {ptsx}\n"
                f"\t- ptsy: {ptsy}\n"
                f"\t- ptsz: {ptsz}\n"
            )
            raise Exception(msg)

        if not (ptsx.shape == ptsy.shape == ptsz.shape):
            msg = (
                "Args (ptsx, ptsy, ptsz) must have the same shape!\n"
                f"\t- ptsx.shape: {ptsx.shape}\n"
                f"\t- ptsy.shape: {ptsy.shape}\n"
                f"\t- ptsz.shape: {ptsz.shape}\n"
            )
            raise Exception(msg)

    # ---------------
    # pfe_in
    # ---------------

    else:

        c0 = (
            isinstance(pfe_in, str)
            and os.path.isfile(pfe_in)
            and (pfe_in.endswith('.csv') or pfe_in.endswith('.dat'))
        )

        if not c0:
            msg = (
                "Arg pfe_in must be a path to a .csv file!\n"
                f"Provided: {pfe_in}"
            )
            raise Exception(msg)
        key = None

    # ---------------
    # outline_only
    # ---------------

    outline_only = ds._generic_check._check_var(
        outline_only, 'outline_only',
        types=bool,
        default=True,
    )

    include_centroid = ds._generic_check._check_var(
        include_centroid, 'include_centroid',
        types=bool,
        default=True,
    )

    # ---------------
    # factor
    # ---------------

    factor = float(ds._generic_check._check_var(
        factor, 'factor',
        types=(float, int),
        default=1.,
    ))

    # ---------------
    # iso
    # ---------------

    iso = 'ISO-10303-21'

    # ---------------
    # pfe_save
    # ---------------

    # Default
    if pfe_save is None:
        path = os.path.abspath('.')
        name = key if key is not None else _NAME
        pfe_save = os.path.join(path, f"{name}.stp")

    # check
    c0 = (
        isinstance(pfe_save, str)
        and (
            os.path.split(pfe_save)[0] == ''
            or os.path.isdir(os.path.split(pfe_save)[0])
        )
    )
    if not c0:
        msg = (
            "Arg pfe_save must be a saving file str ending in '.stp'!\n"
            f"Provided: {pfe_save}"
        )
        raise Exception(msg)

    # makesure extension is included
    if not pfe_save.endswith('.stp'):
        pfe_save = f"{pfe_save}.stp"

    pfe_save = os.path.abspath(pfe_save)

    # ----------------
    # overwrite
    # ----------------

    overwrite = ds._generic_check._check_var(
        overwrite, 'overwrite',
        types=bool,
        default=False,
    )

    return (
        key, key_cam,
        ptsx, ptsy, ptsz,
        pfe_in,
        outline_only,
        include_centroid,
        factor, color,
        iso,
        pfe_save, overwrite,
    )


# #################################################################
# #################################################################
#          extract
# #################################################################


def _extract(
    coll=None,
    key=None,
    key_cam=None,
    ptsx=None,
    ptsy=None,
    ptsz=None,
    outline_only=None,
    include_centroid=None,
    pfe_in=None,
    fname=None,
):

    # ----------------------
    # initialize
    # ----------------------

    dptsx = {}
    dptsy = {}
    dptsz = {}

    # ----------------------
    # extract points from csv
    # ----------------------

    if pfe_in is not None:

        # load csv/dat
        out = np.loadtxt(pfe_in, comments='#')

        # read header
        with open(pfe_in, 'r') as fn:
            header = fn.readline()

        ln = [ss for ss in header[1:].strip().split(' ')]
        if len(ln) != 2:
            msg = (
                "Header format is non-conform (use spaces in header)!\n"
                f"\t- file: {pfe_in}\n"
                f"\t- header: {header}\n"
                "Expected:\n"
                "#   nlos   npts\n(hastag then 2 integers separated by spaces)"
            )
            raise Exception(msg)

        nlos, npts = [int(ss) for ss in ln]

        # safety check
        if out.shape != (nlos * npts, 3):
            msg = (
                f"Arg pfe_in should be a file holding a "
                "(nlos={nlos} x npts={npts}, 3) array"
                f"Provided shape: {out.shape}"
            )
            raise Exception(msg)

        # extract pts as (2, nrays) arrays
        dptsx[fname] = np.array([out[ii::npts, 0] for ii in range(npts)])
        dptsy[fname] = np.array([out[ii::npts, 1] for ii in range(npts)])
        dptsz[fname] = np.array([out[ii::npts, 2] for ii in range(npts)])

    # ----------------------
    # extract points from array
    # ----------------------

    elif ptsx is not None:

        dptsx[fname] = ptsx
        dptsy[fname] = ptsy
        dptsz[fname] = ptsz

    # ----------------------
    # extract points from coll
    # ----------------------

    else:

        if key_cam is None:
            dptsx[fname], dptsy[fname], dptsz[fname] = coll.get_rays_pts(key=key)

        else:
            for kcam in key_cam:
                dptsx[kcam], dptsy[kcam], dptsz[kcam] = coll.get_rays_pts(key=key, key_cam=kcam)

        # ----------
        # outline_only

        if outline_only is True:
            for kcam, v0 in dptsx.items():
                if v0.ndim == 3:
                    iout = ~_get_outline2d(
                        dptsx[kcam][1, ...],
                        include_centroid=include_centroid,
                    )
                    dptsx[kcam][:, iout] = np.nan
                    dptsy[kcam][:, iout] = np.nan
                    dptsz[kcam][:, iout] = np.nan

    return dptsx, dptsy, dptsz


def _get_outline2d(pts, include_centroid=None):

    # --------------
    # eliminate nan

    iok = np.isfinite(pts)
    n0, n1 = pts.shape

    iok2 = np.zeros((n0, n1), dtype=bool)
    tot = np.zeros((n0, n1), dtype=int)

    # ---------------
    # 8-pix filters
    # -----------------

    # -------------------------------------
    # left-side and right-side neighbours

    # left
    iok2[:, 1:] = iok[:, :-1]
    tot += iok2

    # right
    iok2[...] = False
    iok2[:, :-1] = iok[:, 1:]
    tot += iok2

    # top
    iok2[...] = False
    iok2[1:, :] = iok[:-1, :]
    tot += iok2

    # bottom
    iok2[...] = False
    iok2[:-1, :] = iok[1:, :]
    tot += iok2

    # ----------------
    # 4 corners

    # top left
    iok2[...] = False
    iok2[1:, 1:] = iok[:-1, :-1]
    tot += iok2

    # bottom left
    iok2[...] = False
    iok2[:-1, 1:] = iok[1:, :-1]
    tot += iok2

    # top right
    iok2[...] = False
    iok2[1:, :-1] = iok[:-1, 1:]
    tot += iok2

    # bottom right
    iok2[...] = False
    iok2[:-1, :-1] = iok[1:, 1:]
    tot += iok2

    ind = ((tot<7) & iok)

    # ------------------
    # include_centroid

    if include_centroid is True:

        x0 = np.repeat(np.arange(0, n0)[:, None], n1, axis=1)[iok]
        x1 = np.repeat(np.arange(0, n1)[None, :], n0, axis=0)[iok]
        i0 = int(np.round(np.sum(x0) / np.sum(iok)))
        i1 = int(np.round(np.sum(x1) / np.sum(iok)))

        ind[i0, i1] = True

    return ind


# #################################################################
# #################################################################
#           Get color dict
# #################################################################


def _get_dcolor(dptsx=None, color=None):

    # ---------------
    # color
    # ---------------

    if color is None:
        color = _COLOR

    if mcolors.is_color_like(color):

        dcolor = {k0: color for k0 in dptsx.keys()}

    elif isinstance(color, dict):

        c0 = (
            sorted(color.keys()) == sorted(dptsx.keys())
            and all([mcolors.is_color_like(v0) for v0 in color.values()])
        )
        if not c0:
            lstr = [
                f"\t- {k0}: is_color_like {mcolors.is_color_like(dcolor[k0])}"
                for k0 in dptsx.keys()
            ]
            msg = (
                "Arg color must be either a single color, or a dict of colors"
                " with keys:\n"
                + "\n".join(lstr)
            )
            raise Exception(msg)

        dcolor = color

    else:
        msg = f"Arg color must be a color-like value\nProvided: {color}"
        raise Exception(msg)

    #  ------------
    # safety check

    dcolor = {k0: mcolors.to_rgb(v0) for k0, v0 in dcolor.items()}

    return dcolor


# #################################################################
# #################################################################
#          save to stp
# #################################################################


def _save(
    msg=None,
    pfe_save=None,
    overwrite=None,
):

    # -------------
    # check before overwriting

    if os.path.isfile(pfe_save):
        err = "File already exists!"
        if overwrite is True:
            err = f"{err} => overwriting"
            warnings.warn(err)
        else:
            err = f"{err}\nFile:\n\t{pfe_save}"
            raise Exception(err)

    # ----------
    # save

    with open(pfe_save, 'w') as fn:
        fn.write(msg)

    # --------------
    # verb

    msg = f"Saved to:\n\t{pfe_save}"
    print(msg)

    return


# #################################################################
# #################################################################
#          HEADER
# #################################################################


def _get_header(
    fname=None,
    iso=None,
):

    # -------------
    # parameters
    # -------------

    # author
    author = getpass.getuser()

    # timestamp
    t0 = dtm.datetime.now()
    tstr = t0.strftime('%Y-%m-%dT%H:%M:%S-05:00')

    # niso
    niso = iso.split('-')[1]

    # -------------
    # Header
    # -------------

    msg = (
f"""{iso};
HEADER;
/* Generated by software containing ST-Developer
 * from STEP Tools, Inc. (www.steptools.com)
 */
/* OPTION: using custom schema-name function */

FILE_DESCRIPTION(
/* description */ (''),
/* implementation_level */ '2;1');

FILE_NAME(
/* name */ '{fname}.stp',
/* time_stamp */ '{tstr}',
/* author */ ('{author}'),
/* organization */ (''),
/* preprocessor_version */ 'ST-DEVELOPER v18.102',
/* originating_system */ 'SIEMENS PLM Software NX2206.4040',
/* authorisation */ '');\n
"""
    + "FILE_SCHEMA (('AUTOMOTIVE_DESIGN { 1 0 " + f"{niso}" + " 214 3 1 1 1 }'));\n"
    + "ENDSEC;"
    )

    return msg


# #################################################################
# #################################################################
#          Utility
# #################################################################


def _get_k0ind(dind_ok=None, ncum=None, lkcam=None):

    def k0ind(
            ii,
            dind_ok=dind_ok,
            ncum=ncum,
            lkcam=lkcam,
        ):

        icam = np.searchsorted(ncum-1, ii)
        inew = ii - ncum[icam-1] if icam>0 else ii

        return lkcam[icam], tuple([tt[inew] for tt in dind_ok[lkcam[icam]]])

    return k0ind


# #################################################################
# #################################################################
#          DATA
# #################################################################


def _get_data(
    dptsx=None,
    dptsy=None,
    dptsz=None,
    fname=None,
    # options
    dcolor=None,
    # norm
    iso=None,
):

    # -----------
    # nrays
    # -----------

    # vectors
    dvx = {k0: np.diff(v0, axis=0) for k0, v0 in dptsx.items()}
    dvy = {k0: np.diff(v0, axis=0) for k0, v0 in dptsy.items()}
    dvz = {k0: np.diff(v0, axis=0) for k0, v0 in dptsz.items()}

    # dok
    dok = {k0: np.isfinite(v0) for k0, v0 in dvx.items()}

    # length
    dlength = {
        k0: np.sqrt(dvx[k0]**2 + dvy[k0]**2 + dvz[k0]**2)
        for k0 in dptsx.keys()
    }

    # directions
    ddx = {k0: dvx[k0] / dlength[k0] for k0 in dptsx.keys()}
    ddy = {k0: dvy[k0] / dlength[k0] for k0 in dptsx.keys()}
    ddz = {k0: dvz[k0] / dlength[k0] for k0 in dptsx.keys()}

    # shapes
    # dshape_vect = {k0: dvx[k0].shape for k0 in dptsx.keys()}

    dnrays = {k0: v0.sum() for k0, v0 in dok.items()}
    nrays = np.sum([v0 for v0 in dnrays.values()])

    # --------------
    # order of kcam

    lkcam = sorted(dptsx.keys())
    k0ind = _get_k0ind(
        dind_ok={k0: v0.nonzero() for k0, v0 in dok.items()},
        ncum=np.cumsum([dnrays[kcam] for kcam in lkcam]),
        lkcam=lkcam,
    )

    # -----------
    # colors
    # -----------

    colors = sorted(set([v0 for v0 in dcolor.values()]))
    ncol = len(colors)

    # -----------------
    # get index
    # ------------------

    i0 = 31
    dind = {
        'GEOMETRIC_CURVE_SET': {'order': 0},
        'PRESENTATION_LAYER_ASSIGNMENT': {'order': 1},
        'STYLED_ITEM': {
            'order': 2,
            'nn': nrays,
        },
        'PRESENTATION_STYLE_ASSIGNMENT': {
            'order': 3,
            'nn': ncol,
        },
        'CURVE_STYLE': {
            'order': 4,
            'nn': ncol,
        },
        'COLOUR_RGB': {
            'order': 5,
            'nn': ncol,
        },
        'DRAUGHTING_PRE_DEFINED_CURVE_FONT': {
            'order': 6,
            # 'nn': nrays,
        },
        'TRIMMED_CURVE': {
            'order': 7,
            'nn': nrays,
        },
        'LINE': {
            'order': 8,
            'nn': nrays,
        },
        'VECTOR': {
            'order': 9,
            'nn': nrays,
        },
        'AXIS2_PLACEMENT_3D': {'order': 10},
        'DIRECTION0': {
            'order': 11,
            'str': "DIRECTION('',(0.,0.,1.));",
        },
        'DIRECTION1': {
            'order': 12,
            'str': "DIRECTION('',(1.,0.,0.));",
        },
        'DIRECTION': {
            'order': 13,
            'nn': nrays,
        },
        'CARTESIAN_POINT0': {
            'order': 14,
            'str': "CARTESIAN_POINT('',(0.,0.,0.));",
        },
        'CARTESIAN_POINT': {
            'order': 15,
            'nn': nrays,
        },
        'MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION': {'order': 16},
    }

    # complement
    lkey = [k0 for k0 in dind.keys()]
    lorder = [dind[k0]['order'] for k0 in lkey]

    # safety ceck
    assert np.unique(lorder).size == len(lorder)
    inds = np.argsort(lorder)
    lkey = [lkey[ii] for ii in inds]

    # derive indices
    for k0 in lkey:
        nn = dind[k0].get('nn', 1)
        dind[k0]['ind'] = i0 + np.arange(0, nn)
        i0 += nn

    # -----------------
    # COLOUR_RGB
    # -----------------

    k0 = 'COLOUR_RGB'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}('color {ii}',{colors[ii][0]},{colors[ii][1]},{colors[ii][2]});")
    dind[k0]['msg'] = "\n".join(lines)

    # ni = dind[k0]['ind'][0]
    # dind[k0]['msg'] = f"#{ni}={k0}('Medium Royal',{color[0]},{color[1]},{color[2]});"
    # dind[k0]['msg'] = f"#{ni}={k0}('Medium Royal',0.301960784313725,0.427450980392157,0.701960784313725);"

    # -----------------
    # CARTESIAN_POINT
    # -----------------

    k0 = 'CARTESIAN_POINT'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',({dptsx[kcam][ind]},{dptsy[kcam][ind]},{dptsz[kcam][ind]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # DIRECTION
    # -----------------

    k0 = 'DIRECTION'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',({ddx[kcam][ind]},{ddy[kcam][ind]},{ddz[kcam][ind]}));")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # VECTOR
    # -----------------

    k0 = 'VECTOR'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',#{dind['DIRECTION']['ind'][ii]},{dlength[kcam][ind]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # AXIS2_PLACEMENT_3D
    # -----------------

    k0 = 'AXIS2_PLACEMENT_3D'
    ni = dind[k0]['ind'][0]
    lstr = ', '.join([f"#{ii}" for ii in dind['TRIMMED_CURVE']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('',#{dind['CARTESIAN_POINT0']['ind'][0]},#{dind['DIRECTION0']['ind'][0]},#{dind['DIRECTION1']['ind'][0]});"

    # -----------------
    # LINE
    # -----------------

    k0 = 'LINE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',#{dind['CARTESIAN_POINT']['ind'][ii]},#{dind['VECTOR']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # TRIMMED_CURVE
    # -----------------

    k0 = 'TRIMMED_CURVE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',#{dind['LINE']['ind'][ii]},(PARAMETER_VALUE(0.)),(PARAMETER_VALUE(1.)),.T.,.PARAMETER.);")
    dind[k0]['msg'] = "\n".join(lines)

    # ----------------
    # DRAUGHTING_PRE_DEFINED_CURVE_FONT
    # ----------------

    k0 = 'DRAUGHTING_PRE_DEFINED_CURVE_FONT'
    ni = dind[k0]['ind'][0]
    dind[k0]['msg'] = f"#{ni}={k0}('continuous');"

    # ------------------
    # CURVE_STYLE
    # ------------------

    k0 = 'CURVE_STYLE'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}('style {ii}',#{dind['DRAUGHTING_PRE_DEFINED_CURVE_FONT']['ind'][0]},POSITIVE_LENGTH_MEASURE(0.7),#{dind['COLOUR_RGB']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # PRESENTATION_STYLE_ASSIGNMENT
    # ------------------

    k0 = 'PRESENTATION_STYLE_ASSIGNMENT'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        lines.append(f"#{ni}={k0}((#{dind['CURVE_STYLE']['ind'][ii]}));")
    dind[k0]['msg'] = "\n".join(lines)

    #1605=PRESENTATION_STYLE_ASSIGNMENT((#2488));

    # -----------------
    # STYLED_ITEM
    # -----------------

    k0 = 'STYLED_ITEM'
    lines = []
    for ii, ni in enumerate(dind[k0]['ind']):
        kcam, ind = k0ind(ii)
        jj = colors.index(dcolor[kcam])
        lines.append(f"#{ni}={k0}('{kcam}_{ind}',(#{dind['PRESENTATION_STYLE_ASSIGNMENT']['ind'][jj]}),#{dind['TRIMMED_CURVE']['ind'][ii]});")
    dind[k0]['msg'] = "\n".join(lines)

    # -----------------
    # GEOMETRIC_CURVE_SET
    # -----------------

    k0 = 'GEOMETRIC_CURVE_SET'
    ni = dind[k0]['ind'][0]
    lstr = ','.join([f"#{ii}" for ii in dind['TRIMMED_CURVE']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('None',({lstr}));"

    # ----------------------
    # PRESENTATION_LAYER_ASSIGNMENT
    # ----------------------

    k0 = 'PRESENTATION_LAYER_ASSIGNMENT'
    ni = dind[k0]['ind'][0]
    lstr = ','.join([f"#{ii}" for ii in dind['TRIMMED_CURVE']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('1','Layer 1',({lstr}));"

    # ----------------------------------
    # MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION
    # ----------------------------------

    k0 = 'MECHANICAL_DESIGN_GEOMETRIC_PRESENTATION_REPRESENTATION'
    ni = dind[k0]['ind'][0]
    lstr = ','.join([f"#{ii}" for ii in dind['STYLED_ITEM']['ind']])
    dind[k0]['msg'] = f"#{ni}={k0}('',({lstr}),#{i0});"

    # ------------
    # LEFTOVERS
    # ------------

    for k0, v0 in dind.items():
        if v0.get('msg') is None:
            if v0.get('str') is None:
                msg = f"Looks like '{k0}' is missing!"
                raise Exception(msg)
            else:
                ni = dind[k0]['ind'][0]
                dind[k0]['msg'] = f"#{ni}={v0['str']}"


    # --------------------
    # msg_pre
    # --------------------

    msg_pre = (
f"""
DATA;
#10=PROPERTY_DEFINITION_REPRESENTATION(#14,#12);
#11=PROPERTY_DEFINITION_REPRESENTATION(#15,#13);
#12=REPRESENTATION('',(#16),#{i0});
#13=REPRESENTATION('',(#17),#{i0});
#14=PROPERTY_DEFINITION('pmi validation property','',#21);
#15=PROPERTY_DEFINITION('pmi validation property','',#21);
#16=VALUE_REPRESENTATION_ITEM('number of annotations',COUNT_MEASURE(0.));
#17=VALUE_REPRESENTATION_ITEM('number of views',COUNT_MEASURE(0.));
#18=SHAPE_REPRESENTATION_RELATIONSHIP('None', 'relationship between {fname}-None and {fname}-None',#30,#19);
#19=GEOMETRICALLY_BOUNDED_WIREFRAME_SHAPE_REPRESENTATION('{fname}-None',(#31),#{i0});
#20=SHAPE_DEFINITION_REPRESENTATION(#21,#30);
#21=PRODUCT_DEFINITION_SHAPE('','',#22);
#22=PRODUCT_DEFINITION(' ','',#24,#23);
#23=PRODUCT_DEFINITION_CONTEXT('part definition',#29,'design');
#24=PRODUCT_DEFINITION_FORMATION_WITH_SPECIFIED_SOURCE(' ',' ',#26,.NOT_KNOWN.);
#25=PRODUCT_RELATED_PRODUCT_CATEGORY('part','',(#26));
#26=PRODUCT('{fname}','{fname}',' ', (#27));
#27=PRODUCT_CONTEXT(' ',#29,'mechanical');
#28=APPLICATION_PROTOCOL_DEFINITION('international standard','automotive_design',2010,#29);
#29=APPLICATION_CONTEXT('core data for automotive mechanical design processes');
#30=SHAPE_REPRESENTATION('{fname}-None',(#6215),#{i0});
"""
    )

    # --------------------
    # msg_post
    # --------------------

    # 5->91
    ind = i0 + np.arange(0, 8)
    msg_post = (
f"""
#{ind[0]}=(
GEOMETRIC_REPRESENTATION_CONTEXT(3)
GLOBAL_UNCERTAINTY_ASSIGNED_CONTEXT((#{ind[1]}))
GLOBAL_UNIT_ASSIGNED_CONTEXT((#{ind[7]},#{ind[3]},#{ind[2]}))
REPRESENTATION_CONTEXT('{fname}','TOP_LEVEL_ASSEMBLY_PART')
);
#{ind[1]}=UNCERTAINTY_MEASURE_WITH_UNIT(LENGTH_MEASURE(2.E-5),#{ind[7]}, 'DISTANCE_ACCURACY_VALUE','Maximum Tolerance applied to model');
#{ind[2]}=(
NAMED_UNIT(*)
SI_UNIT($,.STERADIAN.)
SOLID_ANGLE_UNIT()
);
#{ind[3]}=(
CONVERSION_BASED_UNIT('DEGREE',#{ind[5]})
NAMED_UNIT(#{ind[4]})
PLANE_ANGLE_UNIT()
);
#{ind[4]}=DIMENSIONAL_EXPONENTS(0.,0.,0.,0.,0.,0.,0.);
#{ind[5]}=PLANE_ANGLE_MEASURE_WITH_UNIT(PLANE_ANGLE_MEASURE(0.0174532925), #{ind[6]});
#{ind[6]}=(
NAMED_UNIT(*)
PLANE_ANGLE_UNIT()
SI_UNIT($,.RADIAN.)
);
#{ind[7]}=(
LENGTH_UNIT()
NAMED_UNIT(*)
SI_UNIT(.MILLI.,.METRE.)
);
ENDSEC;
END-{iso};"""
    )

    # --------------------
    # assemble
    # --------------------

    msg = msg_pre + "\n".join([dind[k0]['msg'] for k0 in lkey]) + msg_post

    return msg


# -*- coding: utf-8 -*-
"""
Created on Thu May 30 16:33:59 2024

@author: dvezinet
"""


import numpy as np


# ##################################################################
# ##################################################################
#                   main
# ##################################################################


def main(
    coll=None,
    key=None,
    key_cam=None,
    data=None,
):

    # ---------------------
    # initialize
    # ---------------------

    static = True
    ddata, dref = {}, {}
    doptics = coll.dobj['diagnostic'][key]['doptics']

    # ---------------------
    # loop on cameras
    # ---------------------

    for cc in key_cam:

        # --------------
        # safety check

        ref = coll.dobj['camera'][cc]['dgeom']['ref']
        dvos = doptics[cc].get('dvos')
        if dvos is None:
            msg = (
                f"Data '{data}' cannot be retrived for diag '{key}' "
                "cam '{cc}' because no dvos computed"
            )
            raise Exception(msg)

        # ----------------
        # vos_pix_sang_sum

        if data == 'vos_pix_sang_integ':

            kdata = dvos['sang_cross']
            ddata[cc] = np.nansum(coll.ddata[kdata]['data'], axis=-1)
            dref[cc] = ref
            units = coll.ddata[kdata]['units']

        # --------------
        # average observation angle per RZ point per pixel in cross section

        elif data == 'vos_cross_ang':

            raise NotImplementedError()

            # solid angle map
            ksang = 'sang_cross'
            ref = coll.ddata[ksang]['ref']
            iok = np.isfinite(coll.ddata[ksang]['data'])

            # pixel centers
            # cr =
            # cz =

            # vect_cross
            ang = np.full(iok.shape, np.nan)
            sli = [None for ss in iok.shape]
            sli[-1] = slice(None)
            sli = tuple(sli)

            # ang[iok] = np.arctan2(
            #    R[sli] - cz[..., None],
            #    Z[sli] - cr[..., None],
            # )

            ddata[cc] = ang
            dref[cc] = ref
            units = coll.ddata[kdata]['units']

    return ddata, dref, units, static


# -*- coding: utf-8 -*-
"""
Created on Wed Jun 12 18:12:43 2024

@author: dvezinet
"""


import warnings


import numpy as np
import astropy.units as asunits
import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import datastock as ds


# ##################################################################
# ##################################################################
#             interpolate along los
# ##################################################################


def main(
    coll=None,
    key_diag=None,
    key_cam=None,
    key_integrand=None,
    key_coords=None,
    # sampling
    res=None,
    mode=None,
    segment=None,
    radius_max=None,
    concatenate=None,
    # interpolating
    domain=None,
    val_out=None,
    # plotting
    vmin=None,
    vmax=None,
    plot=None,
    dcolor=None,
    dax=None,
):

    # ------------
    # check inputs
    # ------------

    (
        key_diag, key_cam, key_los,
        key_integrand, key_coords,
        key_bs_integrand, key_bs_coords,
        concatenate,
        lok_coords, segment, mode, radius_max,
        plot, dcolor,
    ) = _integrate_along_los_check(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        key_integrand=key_integrand,
        key_coords=key_coords,
        # sampling
        segment=segment,
        mode=mode,
        radius_max=radius_max,
        concatenate=concatenate,
        # plotting
        plot=plot,
        dcolor=dcolor,
    )

    # --------------
    # prepare output
    # --------------

    dx = {}
    dy = {}
    dout = {}

    # key_los
    if key_los is not None:
        doptics = {k0: {'los': k0} for k0 in key_los}
    else:
        doptics = coll.dobj['diagnostic'][key_diag]['doptics']

    # ---------------
    # loop on cameras
    # ---------------

    axis_los = 0

    # -----------------------
    # only static coordinates

    if key_integrand in lok_coords and key_coords in lok_coords:

        for ii, kk in enumerate(key_cam):

            klos = doptics[kk]['los']
            if klos is None:
                continue

            # sample
            dx[kk], dy[kk] = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=concatenate,
                return_coords=[key_coords, key_integrand],
            )

    # -------------------------------
    # at least one static coordinate

    elif key_coords in lok_coords or key_integrand in lok_coords:

        # dispatch keys
        if key_coords in lok_coords:
            cll = key_coords
            c2d = key_integrand
            key_bs = key_bs_integrand
        else:
            cll = key_integrand
            c2d = key_coords
            key_bs = key_bs_coords

        for ii, kk in enumerate(key_cam):

            klos = doptics[kk]['los']
            if klos is None:
                continue

            # sample
            pts_x, pts_y, pts_z, pts_ll = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=concatenate,
                return_coords=['x', 'y', 'z', cll],
            )

            Ri = np.hypot(pts_x, pts_y)

            # interpolate
            q2d = coll.interpolate(
                keys=c2d,
                x0=Ri,
                x1=pts_z,
                ref_key=key_bs,
                grid=False,
                domain=domain,
                crop=True,
                nan0=True,
                val_out=val_out,
                return_params=None,
                store=False,
                inplace=False,
            )[c2d]

            # check shapes
            pts_ll, q2d, axis_los = _interpolate_along_los_reshape(
                xdata=pts_ll,
                ydata=q2d['data'],
                yref=q2d['ref'],
            )

            # initialize and fill
            dx[kk] = np.full(q2d.shape, np.nan)
            dy[kk] = np.full(q2d.shape, np.nan)

            isok = np.isfinite(q2d) & np.isfinite(Ri)
            if key_coords in lok_coords:
                dx[kk][isok] = pts_ll[isok]
                dy[kk][isok] = q2d[isok]
            else:
                dx[kk][isok] = q2d[isok]
                dy[kk][isok] = pts_ll[isok]

    # --------------------------------
    # both potentially dynamic 2d data

    else:
        for ii, kk in enumerate(key_cam):

            klos = doptics[kk]['los']
            if klos is None:
                continue

            # sample
            pts_x, pts_y, pts_z = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=concatenate,
                return_coords=['x', 'y', 'z'],
            )

            Ri = np.hypot(pts_x, pts_y)

            # interpolate
            q2dx = coll.interpolate(
                keys=key_coords,
                x0=Ri,
                x1=pts_z,
                ref_key=key_bs_coords,
                grid=False,
                domain=domain,
                crop=True,
                nan0=True,
                val_out=val_out,
                return_params=None,
                store=False,
                inplace=False,
            )[key_coords]['data']

            # interpolate
            q2dy = coll.interpolate_profile2d(
                keys=key_integrand,
                x0=Ri,
                x1=pts_z,
                ref_key=key_bs_integrand,
                grid=False,
                domain=domain,
                crop=True,
                nan0=True,
                val_out=val_out,
                return_params=None,
                store=False,
                inplace=False,
            )[key_integrand]['data']

            # check shape
            if q2dx['data'].shape != q2dy['data'].shape:
                msg = "The two interpolated quantities must have same shape!"
                raise Exception(msg)

            # prepare
            dx[kk] = np.full(q2d.shape, np.nan)
            dy[kk] = np.full(q2d.shape, np.nan)

            # isok
            isok = np.isfinite(q2dx) & np.isfinite(q2dy) & np.isfinite(Ri)
            dx[kk][isok] = q2dx[isok]
            dy[kk][isok] = q2dy[isok]

    # -----------------
    # safety check
    # -----------------

    lout = [k0 for k0, v0 in dx.items() if v0 is None]
    if len(lout) > 0:
        lstr = [f"\t- {k0}" for k0 in lout]
        msg = (
            "The following rays seem to have no existence in desired range:\n"
            + "\n".join(lstr)
            + "\nDetails:\n"
            f"\t- segment: {segment}\n"
            f"\t- radius_max: {radius_max}\n"
        )
        raise Exception(msg)

    # ------------
    # units
    # ------------

    ldist = ['x', 'y', 'z', 'R', 'l', 'ltot']
    lang = ['phi', 'ang_vs_ephi']

    # ---------
    # coords

    if key_coords in lok_coords:
        if key_coords in ldist:
            units_coords = asunits.Unit('m')
        elif key_coords in lang:
            units_coords = asunits.Unit('rad')
        else:
            units_coords = key_coords
    else:
        units_coords = coll.ddata[key_coords]['units']

    # ----------
    # integrand

    if key_integrand in lok_coords:
        if key_integrand in ldist:
            units_integrand = asunits.Unit('m')
        elif key_integrand in lang:
            units_integrand = asunits.Unit('rad')
        else:
            units_integrand = key_integrand
    else:
        units_integrand = coll.ddata[key_integrand]['units']

    # ---------------------
    # add indices
    # ---------------------

    dind = _get_dind(
        coll=coll,
        doptics=doptics,
        dx=dx,
        dy=dy,
    )

    # ------------
    # format
    # ------------

    dout['integrand'] = {
        'key': key_integrand,
        'ddata': dy,
        'units': units_integrand,
    }
    dout['coords'] = {
        'key': key_coords,
        'ddata': dx,
        'units': units_coords,
    }
    dout['ind'] = dind

    # ------------
    # plot
    # ------------

    if plot is True:
        return interpolate_along_los_plot(
            dout=dout,
            axis_los=axis_los,
            # plotting
            vmin=vmin,
            vmax=vmax,
            plot=plot,
            dcolor=dcolor,
            dax=dax,
        )
    else:
        return dout


# ################################################################
#                     Check
# ################################################################


def _integrate_along_los_check(
    coll=None,
    key_diag=None,
    key_cam=None,
    key_integrand=None,
    key_coords=None,
    # sampling
    segment=None,
    mode=None,
    radius_max=None,
    concatenate=None,
    # plotting
    plot=None,
    dcolor=None,
):

    # -----------------
    # keys of diag, cam
    # -----------------

    lrays = list(coll.dobj.get('rays', {}).keys())
    ldiag = list(coll.dobj.get('diagnostic', {}).keys())
    lc = [
            isinstance(key_diag, str)
            and key_diag in lrays
            and key_diag not in ldiag,
            isinstance(key_diag, list)
            and all([
                isinstance(kd, str)
                and kd in lrays
                and kd not in ldiag
                for kd in key_diag
            ]),
    ]

    if any(lc):
        if lc[0]:
            key_los = [key_diag]
        else:
            key_los = key_diag
        key_cam = key_los
    else:
        # key_cam
        key_diag, key_cam = coll.get_diagnostic_cam(
            key=key_diag,
            key_cam=key_cam,
            default='all',
        )
        key_los = None

    # -------------------------
    # keys of coords, integrand
    # -------------------------

    # ---------------------
    # key_data: coordinates

    lok_coords = [
        'x', 'y', 'z', 'R', 'phi', 'ang_vs_ephi',
        'k', 'l', 'ltot', 'itot',
    ]

    # -------------------
    # key_data: integrand

    dp2d = coll.get_profiles2d()
    lok_2d = list(dp2d.keys())
    key_integrand = ds._generic_check._check_var(
        key_integrand, 'key_integrand',
        types=str,
        default='k',
        allowed=lok_coords + lok_2d,
    )

    # --------------------------
    # key_integrand: 2d or not ?

    if key_integrand in lok_2d:
        key_bs_integrand = dp2d[key_integrand]
    else:
        key_bs_integrand = None

    # ------------------
    # key_coords: check

    key_coords = ds._generic_check._check_var(
        key_coords, 'key_coords',
        types=str,
        default='k',
        allowed=lok_coords + lok_2d,
    )

    # -----------------------
    # key_coords: 2d or not ?

    if key_coords in lok_2d:
        key_bs_coords = dp2d[key_coords]
    else:
        key_bs_coords = None

    # -----------------
    # Sampling parameters
    # -----------------

    # segment
    segment = ds._generic_check._check_var(
        segment, 'segment',
        types=int,
        default=-1,
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='abs',
        allowed=['abs', 'rel'],
    )

    # radius_max
    if radius_max is None and mode == 'abs':

        if key_bs_integrand is None and key_bs_coords is None:
            pass

        else:
            wm = coll._which_mesh
            wbs = coll._which_bsplines

            rmax0, rmax1 = 0, 0
            if key_bs_integrand is not None:
                keym = coll.dobj[wbs][key_bs_integrand][wm]
                submesh = coll.dobj[wm][keym]['submesh']
                if submesh is not None:
                    keym = submesh
                knotsR = coll.dobj[wm][keym]['knots'][0]
                rmax0 = np.max(coll.ddata[knotsR]['data'])

            if key_bs_coords is not None:
                keym = coll.dobj[wbs][key_bs_coords][wm]
                submesh = coll.dobj[wm][keym]['submesh']
                if submesh is not None:
                    keym = submesh
                knotsR = coll.dobj[wm][keym]['knots'][0]
                rmax1 = np.max(coll.ddata[knotsR]['data'])

            radius_max = max(rmax0, rmax1)

    # -----------------
    # Plotting parameters
    # -----------------

    # ------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # -----------------
    # concatenate
    # -----------------

    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=plot,
    )

    if plot is True and concatenate is False:
        msg = "Arg concatenate must be True if plot is True"
        raise Exception(msg)

    # --------
    # dcolor
    # --------

    dcolor = _plot_check(
        dcolor=dcolor,
        key_diag=key_diag,
        key_cam=key_cam,
    )

    return (
        key_diag, key_cam, key_los,
        key_integrand, key_coords,
        key_bs_integrand, key_bs_coords,
        concatenate,
        lok_coords, segment, mode, radius_max,
        plot, dcolor,
    )


# ################################################################
#                     reshape
# ################################################################


def _interpolate_along_los_reshape(
    xdata=None,
    ydata=None,
    yref=None,
):

    axis_los = 0
    if xdata.shape != ydata.shape:

        ref = [ss if ss is None else 1 for ss in yref]

        # None matching xdata
        i0 = 0
        for ii in range(ydata.ndim):
            if ref[ii] is None:
                if ydata.shape[ii] == xdata.shape[i0]:
                    ref[ii] = xdata.shape[i0]
                    axis_los = ii
                    i0 += 1
                else:
                    ref[ii] = 1

        # None not macthing xdata (e.g.: domain)
        xdata = xdata.reshape(ref) * np.ones(ydata.shape)

    return xdata, ydata, axis_los


# ################################################################
#                     ind
# ################################################################


def _get_dind(
    coll=None,
    doptics=None,
    dx=None,
    dy=None,
):

    # ----------------
    #  loop on cameras

    dind = {}
    for kcam in doptics.keys():

        klos = doptics[kcam]['los']

        # ------------------
        # preliminary check

        inan = (
            np.isnan(dx[kcam])
            & np.isnan(dy[kcam])
        )

        # ------------------
        # preliminary check

        shape = coll.dobj['rays'][klos]['shape'][1:]
        nnan = np.prod(shape)
        if inan.sum() < nnan:
            msg = (
                f"cam '{kcam}' has unconsistent nb of nans:\n"
                f"\t- shape: {shape}\n"
                f"\t- inan.sum(): {inan.sum()}\n"
                f"\t- nnan: {nnan}\n"
            )
            warnings.warn(msg)
            return None

        # ------------------
        # preliminary check

        i10 = 0
        ind = np.full(dy[kcam].shape, -1, dtype=int)
        for i0, i1 in enumerate(inan.nonzero()[0]):
            ii = np.arange(i10, i1)
            ind[ii] = i0
            i10 = i1 + 1

        # -----------
        # safety check

        lc = [np.any(inan[ind>=0]), (ind == -1).sum() < nnan]
        if  any(lc):
            msg = (
                "Inconsistent nans!\n"
                f"\t- lc: {lc}\n"
                f"\t- ind: {ind}\n"
                f"\t- inan: {inan}\n"
                f"\t- isnan: {inan[ind>=0].sum()}\n"
                f"\t- nnan vs -1: {nnan} vs {(ind == -1).sum()}\n"
            )
            raise Exception(msg)

        dind[kcam] = ind

    return dind


# ################################################################
#                     PLOT
# ################################################################


def interpolate_along_los_plot(
    dout=None,
    axis_los=None,
    # plotting
    vmin=None,
    vmax=None,
    plot=None,
    dcolor=None,
    dax=None,
):


    # -------------
    # check inputs
    # -------------

    dcolor = _plot_check(
        dcolor=dcolor,
        key_diag=None,
        key_cam=list(dout['coords']['ddata'].keys()),
    )

    # ------------
    # prepare
    # -------------

    xlab = f"{dout['coords']['key']} ({dout['coords']['units']})"
    ylab = f"{dout['integrand']['key']} ({dout['integrand']['units']})"

    # -----------
    # make figure
    # -----------

    if dax is None:

        fig = plt.figure()

        ax = fig.add_axes([0.15, 0.1, 0.80, 0.8])

        tit = "LOS-interpolated"
        ax.set_title(tit, size=12, fontweight='bold')
        ax.set_xlabel(xlab)
        ax.set_ylabel(ylab)

        dax = {'main': ax}

    elif isinstance(dax, plt.Axes) or issubclass(dax.__class__, plt.Axes):
        dax = {'main': dax}

    # -------------
    # main
    # -------------

    kax = 'main'
    if dax.get(kax) is not None:
        ax = dax[kax]

        for k0 in dout['coords']['ddata'].keys():

            xx = dout['coords']['ddata'][k0]
            yy = dout['integrand']['ddata'][k0]
            if axis_los != 0:
                xx = xx.swapaxes(0, axis_los)
                yy = yy.swapaxes(0, axis_los)

            ax.plot(
                xx,
                yy,
                label=k0,
                **dcolor[k0],
            )

        ax.legend()

        if vmin is not None:
            ax.set_ylim(bottom=vmin)
        if vmax is not None:
            ax.set_ylim(top=vmax)

    return dout, dax


# #####################
# check inputs
# #####################


def _plot_check(
    dcolor=None,
    key_diag=None,
    key_cam=None,
):

    # ------------------
    # preliminary checks
    # ------------------

    if mcolors.is_color_like(dcolor):
        lc = [dcolor]
    else:
        lc = ['k', 'r', 'g', 'b', 'm', 'c']

    if not isinstance(dcolor, dict):
        dcolor = {
            kk: {
                'color': lc[ii % len(lc)],
                'ls': '-',
                'marker': '.',
                'lw': 1,
            }
            for ii, kk in enumerate(key_cam)
        }

    elif not any([kk in dcolor.keys() for kk in key_cam]):
        dcolor = {
            kk: {
                'color': dcolor.get('color', lc[ii % len(lc)]),
                'ls': dcolor.get('ls', '-'),
                'marker': dcolor.get('marker', '.'),
                'lw': dcolor.get('lw', 1),
                'ms': dcolor.get('ms', 6),
            }
            for ii, kk in enumerate(key_cam)
        }

    # ------------
    # check
    # ------------

    try:
        for ii, kk in enumerate(key_cam):
            dcolor[kk] = {
                'color': dcolor.get(kk, {}).get('color', lc[ii % len(lc)]),
                'ls': dcolor.get(kk, {}).get('ls', '-'),
                'lw': dcolor.get(kk, {}).get('lw', 1),
                'marker': dcolor.get(kk, {}).get('marker', '.'),
                'ms': dcolor.get(kk, {}).get('ms', 6),
            }

    except Exception as err:

        msg = (
            "Arg dcolor must be a dict with camera keys specifying plotting:\n"
            f"\t- camera keys for diag '{key_diag}': {key_cam}\n"
            "\t- expected values: dict of 'color', 'ls', 'lw', 'marker', ...\n"
            "\t{\n"
            "\t\t'color': ...,\n"
            "\t\t'ls': ...,\n"
            "\t\t'lw': ...,\n"
            "\t\t'marker': ...,\n"
            "\t\t'ms': ...,\n"
            "\t}\n"
            "\nProvided:\n{dcolor}\n\n\n"
        ) + str(err)
        raise Exception(msg)

    return dcolor


# -*- coding: utf-8 -*-


import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp
from scipy.spatial import ConvexHull
# import matplotlib.pyplot as plt     # DB


import datastock as ds


from ..geom import CamLOS1D


__all__ = ['compute_los_angles']


# ##############################################################
# ##############################################################
#                       Main
# ##############################################################


def compute_los_angles(
    coll=None,
    key=None,
    # los
    dcompute=None,
    # for storing los
    config=None,
    length=None,
    reflections_nb=None,
    reflections_type=None,
    key_nseg=None,
    # for vos based on los
    res=None,
    compute_vos_from_los=None,
    # overwrite
    overwrite=None,
    **kwdargs,
):

    # ------------
    # check inputs
    # ------------

    key, is2d, compute_vos_from_los = _check(
        coll=coll,
        key=key,
        compute_vos_from_los=compute_vos_from_los,
    )

    # ---------------
    # loop on cameras
    # ---------------

    print(f"\tComputing los for diag '{key}'")
    for ii, (key_cam, v0) in enumerate(dcompute.items()):

        if v0['los_x'] is None or not np.any(np.isfinite(v0['los_x'])):
            continue

        # klos
        klos = f'{key}_{key_cam}_los'

        # ref
        ref = coll.dobj['camera'][key_cam]['dgeom']['ref']

        # ------------
        # add los

        cx2, cy2, cz2 = coll.get_camera_cents_xyz(key=key_cam)

        # check overwrite
        if klos in coll.dobj.get('rays', {}).keys():
            if overwrite is True:
                coll.remove_rays(klos)
            else:
                msg = (
                    f"Rays '{klos}' already exists!\n"
                    "Use overwrite=True to force overwrite!\n"
                )
                raise Exception(msg)

        coll.add_rays(
            key=klos,
            start_x=cx2,
            start_y=cy2,
            start_z=cz2,
            vect_x=v0['los_x'],
            vect_y=v0['los_y'],
            vect_z=v0['los_z'],
            ref=ref,
            diag=key,
            key_cam=key_cam,
            config=config,
            length=length,
            strict=False,
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            key_nseg=key_nseg,
        )

        coll._dobj['diagnostic'][key]['doptics'][key_cam]['los'] = klos

        # ---------------------
        # rough estimate of vos

        if compute_vos_from_los is True:

            if ii == 0:
                print(f"\tComputing vos from los for diag '{key}'")

            _vos_from_los(
                coll=coll,
                key=key,
                key_cam=key_cam,
                v0=v0,
                config=config,
                res=res,
                overwrite=overwrite,
            )

        # ----------------------------------------
        # for spectro => estimate angle variations

        if v0['spectro'] is True:

            _angle_spectro(
                coll=coll,
                key=key,
                key_cam=key_cam,
                v0=v0,
                is2d=is2d,
                ref=ref,
                i0=ii,
                overwrite=overwrite,
            )


# ###########################################################
# ###########################################################
#               check
# ###########################################################


def _check(
    coll=None,
    key=None,
    compute_vos_from_los=None,
):

    # ----------
    # key

    lok = list(coll.dobj.get('diagnostic', {}))
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    is2d = coll.dobj['diagnostic'][key]['is2d']

    # ---------------------
    # compute_vos_from_los

    compute_vos_from_los = ds._generic_check._check_var(
        compute_vos_from_los, 'compute_vos_from_los',
        types=bool,
        default=True,
    )

    return key, is2d, compute_vos_from_los


# ###########################################################
# ###########################################################
#               VOS from LOS
# ###########################################################


def _vos_from_los(
    coll=None,
    key=None,
    key_cam=None,
    v0=None,
    config=None,
    res=None,
    overwrite=None,
):

    # --------------
    # check inputs
    # --------------

    # res
    res = ds._generic_check._check_var(
        res, 'res',
        types=float,
        default=0.01,
        sign='>0',
    )

    # -----------
    # prepare
    # -----------

    # optics
    optics = coll.dobj['diagnostic'][key]['doptics'][key_cam]['optics']
    pinhole = coll.dobj['diagnostic'][key]['doptics'][key_cam]['pinhole']

    # lspectro
    lspectro = [
        oo for oo in optics
        if oo in coll.dobj.get('crystal', {}).keys()
        or oo in coll.dobj.get('grating', {}).keys()
    ]

    dgeom = coll.dobj['camera'][key_cam]['dgeom']
    par = dgeom['parallel']

    if par:
        dx, dy, dz = coll.get_camera_dxyz(
            key=key_cam,
            include_center=True,
        )

    else:
        out0 = coll.ddata[dgeom['outline'][0]]['data']
        out1 = coll.ddata[dgeom['outline'][1]]['data']
        ke0 = dgeom['e0']
        ke1 = dgeom['e1']

    if pinhole is True:
        iref = v0['iref']

    # ----------
    # poly_cross
    # ----------

    lpoly_cross = []
    lpoly_hor = []

    shape0 = v0['cx'].shape
    dphi = np.full(((2,) + shape0), np.nan)
    linds = [range(ss) for ss in shape0]

    for ii, ind in enumerate(itt.product(*linds)):

        if not v0['iok'][ind]:
            lpoly_cross.append(None)
            lpoly_hor.append(None)
            continue

        sli = ind + (slice(None),)
        if pinhole is False:
            iref = v0['iref'][ind]

        if not par:
            e0 = [coll.ddata[kk]['data'][ind] for kk in ke0]
            e1 = [coll.ddata[kk]['data'][ind] for kk in ke1]
            dx = out0 * e0[0] + out1 * e1[0]
            dy = out0 * e0[1] + out1 * e1[1]
            dz = out0 * e0[2] + out1 * e1[2]

        # -----------------------
        # get start / end points

        ptsx, ptsy, ptsz = _get_rays_from_pix(
            coll=coll,
            # start points
            cx=v0['cx'][ind],
            cy=v0['cy'][ind],
            cz=v0['cz'][ind],
            dx=dx,  # np.r_[0],
            dy=dy,  # np.r_[0],
            dz=dz,  # np.r_[0],
            # end points
            x0=np.r_[v0['x0'][sli], v0['cents0'][ind]],
            x1=np.r_[v0['x1'][sli], v0['cents1'][ind]],
            coords=coll.get_optics_x01toxyz(key=optics[iref]),
            lspectro=lspectro,
            config=config,
            # debug
            key=key,
        )

        # ---------
        # sampling

        length = np.sqrt(
            np.diff(ptsx, axis=0)**2
            + np.diff(ptsy, axis=0)**2
            + np.diff(ptsz, axis=0)**2
        )

        npts = int(np.ceil(np.nanmax(length) / res))
        indi = np.linspace(0, 1, npts)
        iok = np.all(np.isfinite(ptsx), axis=0)

        # -------------
        # interpolate

        ptsx = scpinterp.interp1d(
            [0, 1],
            ptsx[:, iok],
            kind='linear',
            axis=0,
        )(indi).ravel()

        ptsy = scpinterp.interp1d(
            [0, 1],
            ptsy[:, iok],
            kind='linear',
            axis=0,
        )(indi).ravel()

        ptsz = scpinterp.interp1d(
            [0, 1],
            ptsz[:, iok],
            kind='linear',
            axis=0,
        )(indi).ravel()

        # -------
        # dphi

        phi = np.arctan2(ptsy, ptsx)
        phimin, phimax = np.nanmin(phi), np.nanmax(phi)
        if phimax - phimin > np.pi:
            phimin, phimax = phimax, phimin + 2.*np.pi
        dphi[(0,) + ind] = phimin
        dphi[(1,) + ind] = phimax

        # -----------
        # poly_cross

        ptsr = np.hypot(ptsx, ptsy)

        convh = ConvexHull(np.array([ptsr, ptsz]).T)
        conv0 = ptsr[convh.vertices]
        conv1 = ptsz[convh.vertices]

        lpoly_cross.append((conv0, conv1))

        # -----------
        # poly_hor

        convh = ConvexHull(np.array([ptsx, ptsy]).T)
        conv0 = ptsx[convh.vertices]
        conv1 = ptsy[convh.vertices]

        lpoly_hor.append((conv0, conv1))

    # ------------------------
    # poly_cross harmonization
    # ------------------------

    lnc = [0 if pp is None else pp[0].size for pp in lpoly_cross]
    lnh = [0 if pp is None else pp[0].size for pp in lpoly_hor]
    nmaxc = np.max(lnc)
    nmaxh = np.max(lnh)

    # ------------
    # prepare

    shc = tuple([nmaxc] + list(shape0))
    shh = tuple([nmaxh] + list(shape0))
    pcross0 = np.full(shc, np.nan)
    pcross1 = np.full(shc, np.nan)
    phor0 = np.full(shh, np.nan)
    phor1 = np.full(shh, np.nan)

    for ii, ind in enumerate(itt.product(*linds)):

        if not v0['iok'][ind]:
            continue

        sli = (slice(None),) + ind

        # --------
        # cross

        if lnc[ii] < nmaxc:
            iextra = np.linspace(0.1, 0.9, nmaxc - lnc[ii])
            indi = np.r_[0, iextra, np.arange(1, lnc[ii])].astype(int)

            pcross0[sli] = scpinterp.interp1d(
                range(lnc[ii]),
                lpoly_cross[ii][0],
                kind='linear',
                axis=0,
            )(indi)

            pcross1[sli] = scpinterp.interp1d(
                range(lnc[ii]),
                lpoly_cross[ii][1],
                kind='linear',
                axis=0,
            )(indi)

        else:
            pcross0[sli] = lpoly_cross[ii][0]
            pcross1[sli] = lpoly_cross[ii][1]

        # --------
        # hor

        if lnh[ii] < nmaxh:
            iextra = np.linspace(0.1, 0.9, nmaxh - lnh[ii])
            indi = np.r_[0, iextra, np.arange(1, lnh[ii])].astype(int)

            phor0[sli] = scpinterp.interp1d(
                range(lnh[ii]),
                lpoly_hor[ii][0],
                kind='linear',
                axis=0,
            )(indi)

            phor1[sli] = scpinterp.interp1d(
                range(lnh[ii]),
                lpoly_hor[ii][1],
                kind='linear',
                axis=0,
            )(indi)

        else:
            phor0[sli] = lpoly_hor[ii][0]
            phor1[sli] = lpoly_hor[ii][1]

    # ----------
    # store
    # ----------

    _vos_from_los_store(
        coll=coll,
        key=key,
        key_cam=key_cam,
        pcross0=pcross0,
        pcross1=pcross1,
        phor0=phor0,
        phor1=phor1,
        dphi=dphi,
        overwrite=overwrite,
    )


def _vos_from_los_store(
    coll=None,
    key=None,
    key_cam=None,
    pcross0=None,
    pcross1=None,
    phor0=None,
    phor1=None,
    dphi=None,
    overwrite=None,
):

    # --------
    # dref

    # keys
    knc = f'{key}_{key_cam}_vos_pc_n'
    knh = f'{key}_{key_cam}_vos_ph_n'

    # dict
    dref = {}
    if pcross0 is not None:
        dref[knc] = {'size': pcross0.shape[0]}
    if phor0 is not None:
        dref[knh] = {'size': phor0.shape[0]}

    # -------------
    # data

    # keys
    kpc0 = f'{key}_{key_cam}_vos_pc0'
    kpc1 = f'{key}_{key_cam}_vos_pc1'
    kph0 = f'{key}_{key_cam}_vos_ph0'
    kph1 = f'{key}_{key_cam}_vos_ph1'

    # reshape for 2d camera
    if coll.dobj['camera'][key_cam]['dgeom']['nd'] == '2d':
        shape0 = coll.dobj['camera'][key_cam]['dgeom']['shape']

        if pcross0 is not None:
            shape = tuple(np.r_[pcross0.shape[0], shape0])
            pcross0 = pcross0.reshape(shape)
            pcross1 = pcross1.reshape(shape)

        if phor0 is not None:
            shape = tuple(np.r_[phor0.shape[0], shape0])
            phor0 = phor0.reshape(shape)
            phor1 = phor1.reshape(shape)

    # ref
    refc = tuple([knc] + list(coll.dobj['camera'][key_cam]['dgeom']['ref']))
    refh = tuple([knh] + list(coll.dobj['camera'][key_cam]['dgeom']['ref']))

    # dict
    ddata = {}

    if pcross0 is not None:
        ddata.update({
            kpc0: {
                'data': pcross0,
                'ref': refc,
                'units': 'm',
                'dim': 'length',
                'quant': 'R',
            },
            kpc1: {
                'data': pcross1,
                'ref': refc,
                'units': 'm',
                'dim': 'length',
                'quant': 'Z',
            },
        })

    if phor0 is not None:
        ddata.update({
            kph0: {
                'data': phor0,
                'ref': refh,
                'units': 'm',
                'dim': 'length',
                'quant': 'X',
            },
            kph1: {
                'data': phor1,
                'ref': refh,
                'units': 'm',
                'dim': 'length',
                'quant': 'Y',
            },
        })

    # ----------
    # store

    # update
    for k0 in dref.keys():
        if k0 in coll.dref.keys():
            if overwrite is True:
                coll.remove_ref(k0, propagate=False)
            else:
                msg = (
                    "Storing vos from los for diag '{key}', camera '{keycam}':\n"
                    f"\t- ref '{k0}' already exists!\n"
                    "\t- Use overwrite=True to force overwriting\n"
                )
                raise Exception(msg)

    for k0 in ddata.keys():
        if k0 in coll.ddata.keys():
            if overwrite is True:
                coll.remove_data(k0, propagate=False)
            else:
                msg = (
                    "Storing vos from los for diag '{key}', camera '{keycam}':\n"
                    f"\t- data '{k0}' already exists!\n"
                    "\t- Use overwrite=True to force overwriting\n"
                )
                raise Exception(msg)

    coll.update(dref=dref, ddata=ddata)

    # add pcross
    doptics = coll._dobj['diagnostic'][key]['doptics']
    doptics[key_cam]['dvos'] = {
        'pcross': None if pcross0 is None else (kpc0, kpc1),
        'phor': None if phor0 is None else (kph0, kph1),
        'dphi': dphi,
    }


# ###########################################################
# ###########################################################
#               get multiple rays
# ###########################################################


def _get_rays_from_pix(
    coll=None,
    kref=None,
    # starting points
    cx=None,
    cy=None,
    cz=None,
    dx=None,
    dy=None,
    dz=None,
    # end points
    x0=None,
    x1=None,
    coords=None,
    lspectro=None,
    config=None,
    # debug
    key=None,
):

    # pixels points (start)
    cxi = cx + dx
    cyi = cy + dy
    czi = cz + dz
    nc = cxi.size

    # end points
    exi, eyi, ezi = coords(x0, x1)
    ne = exi.size

    # final shape
    shape = (2, nc*ne)
    pts_x = np.full(shape, np.nan)
    pts_y = np.full(shape, np.nan)
    pts_z = np.full(shape, np.nan)

    # adjust shapes
    pts_x[0, ...] = np.repeat(cxi, ne)
    pts_y[0, ...] = np.repeat(cyi, ne)
    pts_z[0, ...] = np.repeat(czi, ne)

    vect_x = np.tile(exi, nc) - pts_x[0, ...]
    vect_y = np.tile(eyi, nc) - pts_y[0, ...]
    vect_z = np.tile(ezi, nc) - pts_z[0, ...]

    # ----------
    # diag

    for ii, oo in enumerate(lspectro):

        reflect_ptsvect = coll.get_optics_reflect_ptsvect(oo)
        (
            pts_x[0, ...],
            pts_y[0, ...],
            pts_z[0, ...],
            vect_x, vect_y, vect_z,
            _,
            iok,
        ) = reflect_ptsvect(
            pts_x=pts_x[0, ...],
            pts_y=pts_y[0, ...],
            pts_z=pts_z[0, ...],
            vect_x=vect_x,
            vect_y=vect_y,
            vect_z=vect_z,
            strict=True,
            return_x01=False,
        )

    # ----------
    # config

    # prepare D
    D = np.array([pts_x[0, ...], pts_y[0, ...], pts_z[0, ...]])
    uu = np.array([vect_x, vect_y, vect_z])

    mask = np.all(np.isfinite(uu), axis=0)

    # call legacy code
    cam = CamLOS1D(
        dgeom=(D[:, mask], uu[:, mask]),
        config=config,
        Name='',
        Diag='',
        Exp='',
        strict=True,
    )

    # pin
    pin = cam.dgeom['PkIn']

    pts_x[0, mask] = pin[0, :]
    pts_y[0, mask] = pin[1, :]
    pts_z[0, mask] = pin[2, :]

    # pout
    pout = cam.dgeom['PkOut']

    pts_x[-1, mask] = pout[0, :]
    pts_y[-1, mask] = pout[1, :]
    pts_z[-1, mask] = pout[2, :]

    return pts_x, pts_y, pts_z


# ###########################################################
# ###########################################################
#               Angles variations for spectro
# ###########################################################


def _angle_spectro(
    coll=None,
    key=None,
    key_cam=None,
    v0=None,
    is2d=None,
    ref=None,
    i0=None,
    overwrite=None,
):

    # ------------
    # prepare

    angmin = np.full(v0['cx'].shape, np.nan)
    angmax = np.full(v0['cx'].shape, np.nan)

    kref = v0['optics'][v0['iref']]
    ptsvect = coll.get_optics_reflect_ptsvect(key=kref)
    coords = coll.get_optics_x01toxyz(key=kref)

    # dx, dy, dz = coll.get_camera_dxyz(
    #     key=key_cam,
    #     include_center=True,
    # )

    # ------
    # loop

    if i0 == 0:
        msg = f"\tComputing angles for spectro diag '{key}':\n\t\t- cam '{key_cam}':"
    else:
        msg = f"\t\t- cam '{key_cam}':"
    print(msg)

    # langles = []        # DB
    linds = [range(ss) for ss in v0['cx'].shape]
    for ii, ij in enumerate(itt.product(*linds)):

        # verb
        msg = f"\t\t\tpixel {ii+1} / {v0['cx'].size}"
        end = "\n" if ii == v0['cx'].size - 1 else "\r"
        print(msg, end=end, flush=True)

        if not v0['iok'][ij]:
            continue

        # get 3d coordiantes of points on pixel
        cxi = v0['cx'][ij] # + dx
        cyi = v0['cy'][ij] # + dy
        czi = v0['cz'][ij] # + dz

        nc = 1 # cxi.size

        # slice for x0
        sli = tuple(list(ij) + [slice(None)])

        # get 3d coords of points on crystal
        exi, eyi, ezi = coords(
            v0['x0'][sli],
            v0['x1'][sli],
        )
        ne = exi.size

        # cross
        cxi = np.repeat(cxi, ne)
        cyi = np.repeat(cyi, ne)
        czi = np.repeat(czi, ne)

        # get angles of incidence on crystal
        angles = ptsvect(
            pts_x=cxi,
            pts_y=cyi,
            pts_z=czi,
            vect_x=np.tile(exi, nc) - cxi,
            vect_y=np.tile(eyi, nc) - cyi,
            vect_z=np.tile(ezi, nc) - czi,
            strict=True,
            return_x01=False,
        )[6]

        # Correct for approximation of using
        # the same projected reflection from the center for all
        # ang0 = np.nanmean(angles[:ne])
        # angles[ne:] = ang0 + 0.5*(angles[ne:] - ang0)

        angmin[ij] = np.nanmin(angles[:ne])
        angmax[ij] = np.nanmax(angles[:ne])
        # langles.append(angles)      # DB

    # ddata
    kamin = f'{key}_{key_cam}_amin'
    kamax = f'{key}_{key_cam}_amax'
    ddata = {
        kamin: {
            'data': angmin,
            'ref': ref,
            'dim': 'angle',
            'quant': 'angle',
            'name': 'alpha',
            'units': 'rad',
        },
        kamax: {
            'data': angmax,
            'ref': ref,
            'dim': 'angle',
            'quant': 'angle',
            'name': 'alpha',
            'units': 'rad',
        },
    }

    # update
    for k0 in ddata.keys():
        if k0 in coll.ddata.keys():
            if overwrite is True:
                coll.remove_data(k0, propagate=False)
            else:
                msg = (
                    "Storing vos from los for diag '{key}', camera '{keycam}':\n"
                    f"\t- data '{k0}' already exists!\n"
                    "\t- Use overwrite=True to force overwriting\n"
                )
                raise Exception(msg)

    coll.update(ddata=ddata)

    coll._dobj['diagnostic'][key]['doptics'][key_cam]['amin'] = kamin
    coll._dobj['diagnostic'][key]['doptics'][key_cam]['amax'] = kamax


# -*- coding: utf-8 -*-


import warnings
import itertools as itt


import numpy as np
import matplotlib.colors as mcolors
import datastock as ds


# ################################################################
# ################################################################
#                       Diagnostics
# ################################################################


def _diagnostics_check(
    coll=None,
    key=None,
    doptics=None,
    stack=None,
):

    # ------
    # key
    # ------

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('diagnostic', {}),
        short='d',
        key=key,
    )

    # -----------
    # doptics
    # -----------

    doptics, lap, lfilt, lcryst, lgrat = _check_doptics_basics(
        coll=coll,
        doptics=doptics,
        key=key,
    )

    # -----------------
    # types of camera
    # -----------------

    lcam = list(doptics.keys())
    types = [coll.dobj['camera'][k0]['dgeom']['nd'] for k0 in lcam]

    if len(set(types)) > 1:
        msg = (
            f"diag '{key}': all cameras must be of the same type (1d or 2d)!\n"
            f"\t- cameras: {lcam}\n"
            f"\t- types: {types}"
        )
        raise Exception(msg)

    is2d = types[0] == '2d'

    # -------------------------------------------------
    # check all optics are on good side of each camera
    # -------------------------------------------------

    for cam in lcam:

        dgeom_cam = coll.dobj['camera'][cam]['dgeom']
        last_ref = cam
        last_ref_cls = 'camera'

        is2d = (dgeom_cam['nd'] == '2d')
        parallel = dgeom_cam['parallel']

        # -----------------------------------
        # all optics are common to all pixels

        if doptics[cam]['pinhole'] is True:
            for ii, oo in enumerate(doptics[cam]['optics']):
                last_ref_cls, last_ref = _check_optic(
                    coll=coll,
                    key=key,
                    # optic
                    oo=oo,
                    ocls=doptics[cam]['cls'][ii],
                    cam=cam,
                    is2d=is2d,
                    parallel=parallel,
                    ind_pix=None,
                    dgeom_cam=dgeom_cam,
                    # iterated over
                    last_ref_cls=last_ref_cls,
                    last_ref=last_ref,
                )

        # ------------------------------
        # each pixel has specific optics

        else:

            optics = np.array(doptics[cam]['optics'])
            lcls = np.array(doptics[cam]['cls'])
            paths = doptics[cam]['paths']

            lind = [range(ss) for ss in dgeom_cam['shape']]
            for ind in itt.product(*lind):

                sli = tuple(list(ind) + [slice(None)])
                iop = paths[sli]

                lopi = optics[iop]
                lclsi = lcls[iop]

                for ii, oo in enumerate(lopi):
                    last_ref_cls, last_ref = _check_optic(
                        coll=coll,
                        key=key,
                        # optic
                        oo=oo,
                        ocls=lclsi[ii],
                        cam=cam,
                        is2d=is2d,
                        parallel=parallel,
                        ind_pix=ind,
                        dgeom_cam=dgeom_cam,
                        # iterated over
                        last_ref_cls=last_ref_cls,
                        last_ref=last_ref,
                    )

            # no spectro
            assert last_ref_cls == 'camera', (last_ref_cls, last_ref)

    # -----------------
    # is spectro, PHA
    # -----------------

    spectro, PHA = _get_spectro_PHA(
        coll=coll,
        doptics=doptics,
        key=key,
        lcam=lcam,
    )

    # -----------
    # ispectro
    # -----------

    if spectro:
        for k0, v0 in doptics.items():
            doptics[k0]['ispectro'] = [
                ii for ii, cc in enumerate(v0['cls'])
                if cc in ['grating', 'crystal']
            ]

    # -----------------
    # stack
    # -----------------

    stack = ds._generic_check._check_var(
        stack, 'stack',
        types=str,
        default='horizontal',
        allowed=['horizontal', 'vertical'],
    )

    return key, lcam, doptics, is2d, spectro, stack, PHA


# ##################################################################
# ##################################################################
#                 check doptics basics
# ##################################################################


def _check_doptics_basics(
    coll=None,
    doptics=None,
    key=None,
):

    # --------------------
    # level 0: class
    # --------------------

    if isinstance(doptics, str):
        doptics = {doptics: []}

    if isinstance(doptics, (list, tuple)):
        doptics = {doptics[0]: type(doptics)(doptics[1:])}

    if not isinstance(doptics, dict):
        _err_doptics(
            key=key,
            doptics=doptics,
            extra_msg="\nShould be a dict!\n",
        )

    # ----------
    # check keys

    lcam = list(coll.dobj.get('camera', {}).keys())

    c0 = all([
        (isinstance(k0, str) and k0 in lcam)
        and isinstance(v0, (str, list, tuple, dict))
        for k0, v0 in doptics.items()
    ])

    if not c0:
        _err_doptics(
            key=key,
            doptics=doptics,
            extra_msg="\nSome keys (cam) are not valid!\n",
        )

    # --------------------------
    # start re-arranging as dict

    for k0, v0 in doptics.items():

        if isinstance(v0, str):
            doptics[k0] = {'optics': [v0]}
        elif isinstance(v0, (list, tuple)):
            doptics[k0] = {'optics': v0}
        elif isinstance(doptics.get('optics'), str):
            doptics[k0] = {'optics': [doptics[k0]]}

        c0 = (
            isinstance(doptics[k0].get('optics'), (list, tuple))
            and all([isinstance(k1, str) for k1 in doptics[k0]['optics']])
        )
        if not c0:
            _err_doptics(
                key=key,
                doptics=doptics,
                extra_msg="\n'optics' must be a list or tuple of known optics!\n",
            )

    # --------------------
    # level 1: checking each optics class
    # --------------------

    # -------
    # classes

    lap = list(coll.dobj.get('aperture', {}).keys())
    lfilt = list(coll.dobj.get('filter', {}).keys())
    lcryst = list(coll.dobj.get('crystal', {}).keys())
    lgrat = list(coll.dobj.get('grating', {}).keys())

    lop = lap + lfilt + lcryst + lgrat

    # -----------
    # check names

    dkout = {
        k0: [k1 for k1 in v0['optics'] if k1 not in lop]
        for k0, v0 in doptics.items()
        if any([k1 not in lop for k1 in v0['optics']])
    }

    if len(dkout) > 0:
        _err_doptics(key=key, dkout=dkout, doptics=doptics)

    # -----------------
    # check path matrix
    # -----------------

    doptics2 = {}
    for k0, v0 in doptics.items():

        # -------------------
        # get relevant shapes

        shape_cam = coll.dobj['camera'][k0]['dgeom']['shape']
        noptics = len(v0['optics'])
        shape = tuple(np.r_[shape_cam, noptics])

        # ---------------------------------------
        # no matrix => pinhole or pure collimator

        if v0.get('paths') is None:

            # pinhole camera
            if isinstance(v0['optics'], list):
                pinhole = True
                doptics[k0]['paths'] = None

            # pure collimator camera
            elif isinstance(v0['optics'], tuple) and len(shape_cam) == 1:

                mod = noptics % shape_cam[0]
                if mod != 0:
                    msg = (
                        "Directly providing optics as tuple for collimators:\n"
                        f"\t- diag = '{key}'\n"
                        f"\t- cam = '{k0}'\n"
                        f"\t- shape_cam = {shape_cam}\n"
                        f"\t- noptics = {noptics}\n"
                        f"\t- mod = {mod}\n"
                        "noptics should be a multiple of the nb of sensors!\n"
                    )
                    raise Exception(msg)

                nmult = noptics // shape_cam[0]
                paths = np.zeros((shape_cam[0], noptics), dtype=bool)
                for ii in range(shape_cam[0]):
                    paths[ii, (ii*nmult):(ii+1)*nmult] = True
                doptics[k0]['paths'] = paths

                doptics[k0]['optics'] = list(v0['optics'])

            else:
                emsg = "\nNeither list nor tuple (shape_cam = {shape_cam})!\n"
                _err_doptics(
                    key=key,
                    doptics=doptics,
                    extra_msg=emsg,
                )

        # ------------------------
        # matrix user-provided

        else:

            # check paths is a boolean array of the right shape
            try:
                doptics[k0]['paths'] = np.asarray(v0['paths'], dtype=bool)
                assert doptics[k0]['paths'].shape == shape

            except Exception as err:
                emsg = (
                    f"\n doptics['{k0}']['paths'].shape = "
                    f"{doptics[k0]['paths'].shape} vs {shape}\n"
                )
                err0 = _err_doptics(
                    key=key,
                    doptics=doptics,
                    returnas=True,
                    shape_cam=emsg,
                )
                raise err0 from err

        # -------------------------------------------
        # check if pinhole (all apertures are common)

        if doptics[k0]['paths'] is not None:

            pinhole = np.all(doptics[k0]['paths'])
            if pinhole is True:
                doptics[k0]['paths'] = None

        # ---------
        # get lcls

        lcls = _get_optics_cls(
            coll=coll,
            optics=doptics[k0]['optics'],
        )[1]

        # ---------------------
        # populate doptics2

        doptics2[k0] = {
            'camera': k0,
            'optics': doptics[k0]['optics'],
            'cls': lcls,
            'pinhole': bool(pinhole),
            'paths': doptics[k0]['paths'],
            'los': None,
            'etendue': None,
            'etend_type': None,
            'amin': None,
            'amax': None,
        }

    return doptics2, lap, lfilt, lcryst, lgrat


# -----------
# raise err
# -----------


def _err_doptics(
    key=None,
    dkout=None,
    doptics=None,
    returnas=None,
    extra_msg=None,
):

    # ---------------
    # msg

    msg = (
        f"diag '{key}': arg doptics must be a dict with:\n"
        "\t- keys: key to existing camera\n"
        "\t- values: one of the following:\n\n"
        "\t\t- list of keys to apertures\n"
        "\t\t\tIn this case tofu will assume pinhole camera\n"
        "\t\t\tMeaning all apertures are common to all pixels\n"
        "\t\t- tuple of keys to aperures\n"
        "\t\t\tIn this case tofu will assume collimator camera\n"
        "\t\t\tMeaning each pixel is associated to N apertures\n"
        "\t\t\tWhere N is an integer napertures = N x ncam\n"
        "\t\t\t(the first N apertures go to the first pixel...)\n"
        "\t\t- (most general) a dict of the form:\n"
        "\t\t\t'optics': list of apertures\n"
        "\t\t\t'paths': (shape_cam, noptics) bool array\n"
    )
    if dkout is not None and len(dkout) > 0:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dkout.items()]
        msg += "Wrong key / value pairs:\n" + "\n".join(lstr)

    if extra_msg is not None:
        msg += extra_msg

    msg += f"\nProvided:\n{doptics}"

    # ---------------
    # raise vs return

    if returnas is True:
        return Exception(msg)
    else:
        raise Exception(msg)


# ##################################################################
# ##################################################################
#                 check optics 3d vs camera
# ##################################################################


def _check_optic(
    coll=None,
    key=None,
    # optic
    oo=None,
    ocls=None,
    cam=None,
    is2d=None,
    parallel=None,
    ind_pix=None,
    dgeom_cam=None,
    # iterated over
    last_ref_cls=None,
    last_ref=None,
):

    # --------------
    # prepare
    # --------------

    dgeom = coll.dobj[ocls][oo]['dgeom']
    px, py, pz = coll.get_optics_poly(key=oo)
    dgeom_lastref = coll.dobj[last_ref_cls][last_ref]['dgeom']

    # -------------------
    # single cent and nin
    # -------------------

    if (last_ref == cam and is2d) or last_ref != cam:

        cent = dgeom_lastref['cent']
        nin = dgeom_lastref['nin']

        iout = (
            (px - cent[0])*nin[0]
            + (py - cent[1])*nin[1]
            + (pz - cent[2])*nin[2]
        ) <= 0

        if np.any(iout):
            msg = (
                f"diag '{key}':\n"
                f"The following points of {ocls} '{oo}' are on the wrong"
                f"side of lastref {last_ref_cls} '{last_ref}':\n"
                f"{iout.nonzero()[0]}\n\n"
                f"last ref {last_ref_cls} '{last_ref}':\n{dgeom_lastref}\n\n"
                f"optics {ocls} '{oo}':\n{dgeom}\n\n"
                "Tip:\n"
                "\tMake sure to provide optics ordered from camera to plasma"
            )
            raise Exception(msg)

    # --------------
    # prepare
    # --------------

    else:

        # --------------------
        # get pixel center(s)

        cx, cy, cz = dgeom_cam['cents']
        if ind_pix is None:
            cx = coll.ddata[cx]['data'][None, ...]
            cy = coll.ddata[cy]['data'][None, ...]
            cz = coll.ddata[cz]['data'][None, ...]

        else:
            cx = coll.ddata[cx]['data'][ind_pix]
            cy = coll.ddata[cy]['data'][ind_pix]
            cz = coll.ddata[cz]['data'][ind_pix]

        # ------------------------
        # get pixel unit vector(s)

        if dgeom_cam['parallel']:
            ninx, niny, ninz = dgeom_cam['nin']
        else:
            ninx, niny, ninz = dgeom_cam['nin']
            if ind_pix is None:
                ninx = coll.ddata[ninx]['data'][None, ...]
                niny = coll.ddata[niny]['data'][None, ...]
                ninz = coll.ddata[ninz]['data'][None, ...]
            else:
                ninx = coll.ddata[ninx]['data'][ind_pix]
                niny = coll.ddata[niny]['data'][ind_pix]
                ninz = coll.ddata[ninz]['data'][ind_pix]

        # --------------------------
        # spot points out of domain

        iout = (
            (px[:, None] - cx)*ninx
            + (py[:, None] - cy)*niny
            + (pz[:, None] - cz)*ninz
        ) <= 0

        # ---------------------
        # raise warning if any

        if np.any(iout):
            msg = (
                f"diag '{key}':\n"
                f"The following points of {ocls} '{oo}' are on the wrong"
                f"side of lastref {last_ref_cls} '{last_ref}':\n"
                f"{iout.nonzero()[0]}\n\n"
                f"last ref {last_ref_cls} '{last_ref}':\n{dgeom_lastref}\n\n"
                f"optics {ocls} '{oo}':\n{dgeom}\n\n"
                "Tip:\n"
                "\tMake sure to provide optics ordered from camera to plasma\n"
            )
            warnings.warn(msg)

    # --------------
    # update last_ref ?
    # --------------

    if ocls in ['crystal', 'grating']:
        last_ref = oo
        last_ref_cls = ocls

    return last_ref_cls, last_ref


# ##################################################################
# ##################################################################
#                Get spectro PHA
# ##################################################################


def _get_spectro_PHA(
    coll=None,
    doptics=None,
    key=None,
    lcam=None,
):

    # --------------
    # is spectro
    # --------------

    dspectro = {
        k0: (
            any([
                k1 in coll.dobj.get('crystal', {}).keys()
                or k1 in coll.dobj.get('grating', {}).keys()
                for k1 in v0['optics']
            ])
        )
        for k0, v0 in doptics.items()
    }

    lc = [
        all([v0 for v0 in dspectro.values()]),
        all([not v0 for v0 in dspectro.values()]),
    ]
    if np.sum(lc) != 1:
        msg = (
            f"diag '{key}' must be either all spectro or all non-spectro!\n"
            + "\n".join([f"\t- {k0}: {v0}" for k0, v0 in dspectro.items()])
        )
        raise Exception(msg)

    spectro = lc[0] is True

    # --------
    # is PHA
    # --------

    dPHA = {
        k0 : coll.dobj['camera'][k0]['dmat']['mode'] == 'PHA'
        for k0 in lcam
        if coll.dobj['camera'][k0].get('dmat') is not None
    }

    if len(dPHA) > 0:
        lc = [
            all([v0 for v0 in dPHA.values()]),
            all([not v0 for v0 in dPHA.values()]),
        ]

        if np.sum(lc) != 1:
            msg = (
                f"diag '{key}' must be either all PHA or all non-PHA!\n"
                + "\n".join([f"\t- {k0}: {v0}" for k0, v0 in dPHA.items()])
            )
            raise Exception(msg)

        PHA = lc[0]
    else:
        PHA = False

    return spectro, PHA


# ##################################################################
# ##################################################################
#                           check
# ##################################################################


def _diagnostics(
    coll=None,
    key=None,
    doptics=None,
    stack=None,
    **kwdargs,
):

    # ------------
    # check inputs

    (
        key, lcam, doptics, is2d, spectro, stack, PHA,
    ) = _diagnostics_check(
        coll=coll,
        key=key,
        doptics=doptics,
        stack=stack,
    )

    # --------
    # dobj

    dobj = {
        'diagnostic': {
            key: {
                'camera': lcam,
                'doptics': doptics,
                'ncam': len(doptics),
                # 'npix tot.': np.sum(),
                'is2d': is2d,
                'spectro': spectro,
                'PHA': PHA,
                'stack': stack,
                'signal': None,
            },
        },
    }

    # -----------
    # kwdargs

    if len(kwdargs) > 0:
        for k0, v0 in kwdargs.items():
            if not isinstance(k0, str):
                continue
            elif k0 in dobj['diagnostic'][key].keys():
                continue
            else:
                dobj['diagnostic'][key][k0] = v0

    return None, None, dobj


# ##################################################################
# ##################################################################
#                           get ref
# ##################################################################


def _get_default_cam(coll=None, key=None, key_cam=None, default=None):

    # ----------
    # key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    is2d = coll.dobj['diagnostic'][key]['is2d']
    # spectro = coll.dobj['diagnostic'][key]['spectro']

    # ---------------
    # default

    default = ds._generic_check._check_var(
        default, 'default',
        types=str,
        default='first' if is2d else 'all',
        allowed=['all', 'first'],
    )

    # -----------------------
    # key_cam (only 1 if is2d)

    lok = list(coll.dobj['diagnostic'][key]['doptics'].keys())
    if default == 'first':
        # 2d: can only select one camera at a time
        key_cam_def = [coll.dobj['diagnostic'][key]['camera'][0]]
    else:
        key_cam_def = None

    if isinstance(key_cam, str):
        key_cam = [key_cam]

    key_cam = ds._generic_check._check_var_iter(
        key_cam, 'key_cam',
        types=list,
        types_iter=str,
        allowed=lok,
        default=key_cam_def,
    )

    return key, key_cam


def get_ref(coll=None, key=None):

    # ---------
    # check key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -----------
    # return

    cam = coll.dobj['diagnostic'][key]['optics'][0]
    return coll.dobj['camera'][cam]['dgeom']['ref']


# ################################################################
# ################################################################
#                      get optics
# ################################################################


def _get_optics_cls(coll=None, optics=None):

    # ---------
    # check key

    lcls = ['camera', 'aperture', 'filter', 'crystal', 'grating']

    if isinstance(optics, str):
        optics = [optics]

    lok = list(itt.chain.from_iterable([
        list(coll.dobj.get(cc, {}).keys())
        for cc in lcls
    ]))

    # check
    optics = ds._generic_check._check_var_iter(
        optics, 'optics',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    # -----------
    # optics_cls

    derr = {}
    optics_cls = []
    for ii, oo in enumerate(optics):
        lc = [cc for cc in lcls if oo in coll.dobj.get(cc, {}).keys()]
        if len(lc) == 1:
            optics_cls.append(lc[0])
        else:
            derr[oo] = lc

    # --------
    # error

    if len(derr) > 0:
        msg = (
            "The following have no / several classes:\n"
            + "\n".join([f"\t- {k0}: {v0}" for k0, v0 in derr.items()])
        )
        raise Exception(msg)

    return optics, optics_cls


# def _get_diagnostic_doptics(coll=None, key=None):

#     # ---------
#     # check key

#     lok = list(coll.dobj.get('diagnostic', {}).keys())
#     key = ds._generic_check._check_var(
#         key, 'key',
#         types=str,
#         allowed=lok,
#     )

#     # ------------------
#     # optics and classes

#     doptics = {}
#     for k0, v0 in coll.dobj['diagnostic'][key]['doptics']:
#         doptics[k0]['camera'] = k0
#         doptics[k0]['optics'], doptics[k0]['cls'] = _get_optics_cls(
#             coll=coll,
#             optics=v0,
#         )

#     # -----------
#     # ispectro

#     if coll.dobj['diagnostic'][key]['spectro']:
#         for k0, v0 in doptics.items():
#             doptics[k0]['ispectro'] = [
#                 ii for ii, cc in enumerate(v0['cls'])
#                 if cc in ['grating', 'crystal']
#             ]

#     return doptics


# ##################################################################
# ##################################################################
#                           set color
# ##################################################################


def _set_optics_color(
    coll=None,
    key=None,
    color=None,
):

    # ------------
    # check inputs

    # key
    lk = ['aperture', 'filter', 'crystal', 'grating', 'camera', 'diagnostic']
    dk = {
        k0: list(coll.dobj.get(k0, {}).keys())
        for k0 in lk
    }
    lok = itt.chain.from_iterable([vv for vv in dk.values()])

    if isinstance(key, str):
        key = [key]

    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    # color
    if color is None:
        color = 'k'

    if not mcolors.is_color_like(color):
        msg = (
            f"Arg color for '{key}' must be a matplotlib color!\n"
            f"Provided: {color}\n"
        )
        raise Exception(msg)

    color = mcolors.to_rgba(color)

    # ------------
    # set color

    for k0 in key:
        cls = [k1 for k1, v1 in dk.items() if k0 in v1][0]
        if cls == 'diagnostic':
            for k2 in coll._dobj[cls][k0]['optics']:
                cls2 = [k1 for k1, v1 in dk.items() if k2 in v1][0]
                coll._dobj[cls2][k2]['dmisc']['color'] = color
        else:
            coll._dobj[cls][k0]['dmisc']['color'] = color

    return


# ##################################################################
# ##################################################################
#                           remove
# ##################################################################


def _remove(
    coll=None,
    key=None,
    key_cam=None,
):

    # ------------
    # check inputs

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    doptics = coll.dobj['diagnostic'][key]['doptics']

    if key_cam is None:
        key_cam = coll.dobj['diagnostic'][key]['camera']

    else:
        if isinstance(key_cam, str):
            key_cam = [key_cam]

        key_cam = ds._generic_check._check_var_iter(
            key_cam, 'key_cam',
            types=list,
            types_iter=str,
            allowed=coll.dobj['diagnostic'][key]['camera'],
        )

    # ------------
    # list data

    lkd = ['etendue']
    ld = []
    for k0 in lkd:
        for k1 in key_cam:
            if doptics[k1][k0] is not None:
                ld.append(doptics[k1][k0])

    # -----------
    # remove data

    if len(ld) > 0:
        coll.remove_data(ld, propagate=True)

    # ----------
    # remove los

    for k1 in key_cam:
        if doptics[k1]['los'] is not None:
            coll.remove_rays(key=doptics[k1]['los'])

    # -----------
    # remove diag

    if len(key_cam) == len(doptics):
        del coll._dobj['diagnostic'][key]


# -*- coding: utf-8 -*-


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import datastock as ds


__all__ = ['spectral_range_2d']


# #################################################################
# #################################################################
#               Main
# #################################################################


def spectral_range_2d(
    # crystal
    lamb0=None,
    bragg0=None,
    # geometry basis
    ap=None,
    ex=None,
    ey=None,
    beta_max=None,
    # geometry
    xx=None,
    length=None,
    rcurve=None,
    varrad_b=None,
    dist=None,
    # camera
    dcam=None,
    # options
    npts=None,
    # plotting
    plot=None,
    dax=None,
    # saving
    save=None,
    pfe_fig=None,
    pfe_npz=None,
):

    """

    lamb0: target wavelength
    bragg0: target bragg angle
    rcurve: radii of curvature

    ap: point source position
    xx: distance between point source and crystals
    dist: lenght of rays after reflexion
    beta_max: maximum angular opening from point source (optionnal)
    npts: nb of rays from point source to crystals
    length: crystal length

    """

    # --------
    # check

    din, npts, plot, save, pfe_fig, pfe_npz = _check(**locals())

    # --------------
    # compute

    crystx, crysty, endx, endy, lamb = _compute(
        npts=npts,
        beta_max=beta_max,
        dcam=dcam,
        **din,
    )

    # -------------
    # format output


    ilamb_min = np.full((lamb.shape[1],), -1)
    ilamb_max = np.full((lamb.shape[1],), -1)
    iok = np.any(np.isfinite(lamb), axis=0)
    ilamb_min[iok] = np.nanargmin(lamb[:, iok], axis=0)
    ilamb_max[iok] = np.nanargmax(lamb[:, iok], axis=0)

    lamb_min = np.array([
        lamb[imin, ii] if imin >= 0 else np.nan
        for ii, imin in enumerate(ilamb_min)
    ])
    lamb_max = np.array([
        lamb[imax, ii] if imax >= 0 else np.nan
        for ii, imax in enumerate(ilamb_max)
    ])

    dout = dict(din)
    dout.update({
        'beta_max': beta_max,
        'crystx': crystx,
        'crysty': crysty,
        'endx': endx,
        'endy': endy,
        'lamb': lamb,
        'ilamb_min': ilamb_min,
        'ilamb_max': ilamb_max,
        'lamb_min': lamb_min,
        'lamb_max': lamb_max,
        'Dlamb': lamb_max - lamb_min,
    })

    if dcam is not None:
        dout['dcam'] = dcam

    # ---------
    # plot

    if plot is True:
        dax = _plot(
            dax=dax,
            pfe_fig=pfe_fig,
            **dout,
        )

    # ----------
    # save

    if save is True:
        np.savez(pfe_npz, **dout)

    # ---------
    # return

    if plot is True:
        return dout, dax
    else:
        return dout


# #################################################################
# #################################################################
#               Check
# #################################################################


def _check(
    # crystal
    lamb0=None,
    bragg0=None,
    # geometry basis
    ap=None,
    ex=None,
    ey=None,
    # geometry
    xx=None,
    length=None,
    rcurve=None,
    varrad_b=None,
    dist=None,
    # options
    npts=None,
    # plotting
    plot=None,
    ax=None,
    # saving
    save=None,
    pfe_fig=None,
    pfe_npz=None,
    # unused
    **kwdargs,
):

    # --------------
    # geometry basis

    basis_def = {'ap': np.r_[0, 0], 'ex': np.r_[1, 0], 'ey': np.r_[0, 1]}
    din_basis = {'ap': ap, 'ex': ex, 'ey': ey}
    for k0, v0 in din_basis.items():

        if v0 is None:
            din_basis[k0] = basis_def[k0]

        din_basis[k0] = np.atleast_1d(din_basis[k0]).ravel().astype(float)[:2]

    # normalize ex
    din_basis['ex'] = din_basis['ex'] / np.linalg.norm(din_basis['ex'])

    # perpendicular + normalize ey
    sca = np.sum(din_basis['ex']*din_basis['ey'])
    din_basis['ey'] = din_basis['ey'] - sca * din_basis['ex']
    din_basis['ey'] = din_basis['ey'] / np.linalg.norm(din_basis['ey'])

    # -----------------
    # initialize dict

    din = {
        'lamb0': lamb0,
        'bragg0': bragg0,
        # geometry
        'xx': xx,
        'length': length,
        'rcurve': rcurve,
        'varrad_b': varrad_b,
        'dist': dist,
    }

    # -------------
    # get size

    # make all arrays
    for k0, v0 in din.items():
        if v0 is None:
            din[k0] = np.nan
        din[k0] = np.atleast_1d(din[k0]).ravel().astype(float)

    # sizes
    lsizes = list(set([v0.size for v0 in din.values()]))
    if len(lsizes) == 1:
        pass
    elif 1 in lsizes and len(lsizes) == 2:
        size = [ss for ss in lsizes if ss != 1][0]
        for k0, v0 in din.items():
            if v0.size == 1:
                din[k0] = np.full((size,), v0[0])
    else:
        lstr = [f"\t- '{k0}': {v0.size}" for k0, v0 in din.items()]
        msg = (
            "All args must be either scalar or 1d arrays of the same size:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    # -------
    # values

    for k0, v0 in din.items():
        if k0 not in ['rcurve', 'varrad_b']:
            c0 = np.all(np.isfinite(v0)) and np.all(v0 >= 0.)

            if not c0:
                msg = (
                    f"Arg '{k0}' must be finite and positive\n"
                    f"Provided: {v0}"
                )
                raise Exception(msg)

    # ------------
    # add basis

    din.update(din_basis)

    # ---------
    # npts

    if npts is None:
        npts = 101
    npts = int(npts)
    if npts % 2 == 0:
        npts += 1

    # ---------
    # plot

    # plot
    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # ---------
    # save

    # save
    save = ds._generic_check._check_var(
        save, 'save',
        types=bool,
        default=False,
    )

    return din, npts, plot, save, pfe_fig, pfe_npz


# #################################################################
# #################################################################
#               Compute
# #################################################################


def _compute(
    # crystal
    lamb0=None,
    bragg0=None,
    # geometry basis
    ap=None,
    ex=None,
    ey=None,
    beta_max=None,
    # geometry
    xx=None,
    length=None,
    rcurve=None,
    varrad_b=None,
    dist=None,
    # options
    npts=None,
    # camera
    dcam=None,
):

    # ------------
    # initialize

    size = lamb0.size

    crystx = np.full((npts, size), np.nan)
    crysty = np.full((npts, size), np.nan)
    vnx = np.full((npts, size), np.nan)
    vny = np.full((npts, size), np.nan)

    # ----------------
    # compute geometry
    # ----------------

    # 2d
    d2 = lamb0 / np.sin(bragg0)

    # summit of crystal
    sx = ap[0] + xx * ex[0]
    sy = ap[1] + xx * ex[1]

    # ------------------------
    # indices of crystal types

    # variable-radii sinusoidal spiral
    indb = np.isfinite(varrad_b)

    # indices of curved crystals
    indc = np.isfinite(rcurve) & (~indb)

    # flat
    indf = (~indc) & (~indb)

    # safety check
    if not np.all(np.sum([indc, indb, indf], axis=0) == 1):
        msg = (
            "Some undetermined 2d crystal shapes:\n"
            f"\t- indc = {indc}\n"
            f"\t- indb = {indb}\n"
            f"\t- indf = {indf}\n"
        )
        raise Exception(msg)

    # ---------------------
    # curved crystals

    # center of curvature
    ecx = np.sin(bragg0[indc]) * ex[0] - np.cos(bragg0[indc]) * ey[0]
    ecy = np.sin(bragg0[indc]) * ex[1] - np.cos(bragg0[indc]) * ey[1]
    ecx_p = -ecy
    ecy_p = ecx

    cx = sx[indc] - rcurve[indc] * ecx
    cy = sy[indc] - rcurve[indc] * ecy

    # half angular opening of crystal
    dalpha = 0.5*length[indc] / rcurve[indc]
    theta = dalpha * np.linspace(-1, 1, npts)[:, None]

    # crystal plotting - curved
    ethetax = np.cos(theta) * ecx[None, :] + np.sin(theta) * ecx_p[None, :]
    ethetay = np.cos(theta) * ecy[None, :] + np.sin(theta) * ecy_p[None, :]

    crystx[:, indc] = cx[None, :] + rcurve[indc][None, :] * ethetax
    crysty[:, indc] = cy[None, :] + rcurve[indc][None, :] * ethetay

    # local normal vectors
    vnx[:, indc] = -ethetax
    vny[:, indc] = -ethetay

    # -----------------------
    # flat crystals

    # crystal plotting - straight
    estraightx = np.cos(bragg0)[indf] * ex[0] + np.sin(bragg0)[indf] * ey[0]
    estraighty = np.cos(bragg0)[indf] * ex[1] + np.sin(bragg0)[indf] * ey[1]

    ll = 0.5 * length[None, indf] * np.linspace(-1, 1, npts)[:, None]
    crystx[:, indf] = sx[None, indf] + ll*estraightx[None, :]
    crysty[:, indf] = sy[None, indf] + ll*estraighty[None, :]

    # local normal vectors
    vnx[:, indf] = -estraighty
    vny[:, indf] = estraightx

    # -----------------------
    # variable radii crystals

    # main parameters
    gam0 = bragg0[indb]
    r0 = rcurve[indb]
    ix = ~np.isfinite(r0)
    r0[ix] = xx[indb][ix]
    b = varrad_b[indb]

    # local radius of curvature at center
    # rc0 = r0 / (b * np.sin(gam0))

    # dOMx = r / (b-1) * (cos(phi) / tan(gam) - sin(phi))
    # dOMy = r / (b-1) * (sin(phi) / tan(gam) + cos(phi))
    # dL = r/(b-1) * 1 / sin(gam)
    # dL ~ r0/(b-1) * 1/sin(gam0) * Dgam

    # half angular opening of crystal (approximative)
    # dgam = 0.5*length / rc0
    dgam = 1.1 * length[indb] * np.sin(gam0) * (b-1) / r0 / 2

    # gam
    gam = gam0[None, :] + dgam[None, :] * np.linspace(-1, 1, npts)[:, None]

    # r
    r = r0[None, :] * (np.sin(gam) / np.sin(gam0)[None, :])**(1/(b[None, :]-1))

    # phi
    phi = (gam - gam0[None, :]) / (b[None, :]-1)

    # cryst

    crystx[:, indb] = (
        ap[0]
        + (xx[indb] - r0) * ex[0]
        + r * (np.cos(phi) * ex[0]  + np.sin(phi) * ey[0])
    )
    crysty[:, indb] = (
        ap[1]
        + (xx[indb] - r0) * ex[1]
        + r * (np.cos(phi) * ex[1]  + np.sin(phi) * ey[1])
    )

    # derivative
    c0 = r / (b[None, :] - 1)
    c1 = np.cos(gam) / np.sin(gam)
    dOMxx = c0 * (c1 * np.cos(phi) - np.sin(phi))
    dOMyy = c0 * (c1 * np.sin(phi) + np.cos(phi))
    dOMx = dOMxx * ex[0] + dOMyy * ey[0]
    dOMy = dOMxx * ex[1] + dOMyy * ey[1]

    # local normal vectors
    vnx[:, indb] = dOMy / np.sqrt(dOMx**2 + dOMy**2)
    vny[:, indb] = -dOMx / np.sqrt(dOMx**2 + dOMy**2)

    # ----------------
    # compute rays
    # ----------------

    # vectors of incident rays
    vix = crystx - ap[0]
    viy = crysty - ap[1]
    vin = np.sqrt(vix**2 + viy**2)
    vix = vix / vin
    viy = viy / vin


    # reflected vectors
    sca = vix*vnx + viy*vny
    vrx = vix - 2.*sca*vnx
    vry = viy - 2.*sca*vny

    # end of rays at dist
    endx = crystx + dist * vrx
    endy = crysty + dist * vry

    # ----------------------
    # compute spectral range

    # get local bragg angle - top and bottom
    bragg = np.arccos(sca) - np.pi/2.

    # lamb
    lamb = d2 * np.sin(bragg)

    # beta_max
    if beta_max is not None:
        dvx, dvy = crystx - ap[0], crysty - ap[1]
        beta = np.arctan2(dvx*ey[0] + dvy*ey[1], dvx*ex[0] + dvy*ex[1])
        ind = np.abs(beta) > beta_max
        endx[ind] = np.nan
        endy[ind] = np.nan
        lamb[ind] = np.nan

    # -----------------
    # impacts on camera
    # -----------------

    if dcam is not None:
        ninx, niny = dcam['nin'][:2]
        ninn = np.sqrt(ninx**2 + niny**2)
        ninx, niny = ninx/ninn, niny/ninn

        ninx_r = ninx * ex[0] + niny * ey[0]
        niny_r = ninx * ex[1] + niny * ey[1]

        camx = ap[0] + dcam['cent'][0] * ex[0] + dcam['cent'][1] * ey[0]
        camy = ap[1] + dcam['cent'][0] * ex[1] + dcam['cent'][1] * ey[1]

        sca_up = (camx - crystx) * ninx_r + (camy - crysty) * niny_r
        sca_bot = vrx*ninx_r + vry*niny_r

        kk = sca_up / sca_bot
        ptsx = crystx + kk * vrx
        ptsy = crysty + kk * vry

        e0x = -niny_r
        e0y = ninx_r
        x0 = (ptsx - camx) * e0x + (ptsy - camy) * e0y

        if beta_max is not None:
            x0[ind] = np.nan

        dcam['x0'] = x0
        dcam['cent_r'] = np.r_[camx, camy]
        dcam['nin_r'] = np.r_[ninx_r, niny_r]

    return crystx, crysty, endx, endy, lamb


# #################################################################
# #################################################################
#               Plot
# #################################################################


def _plot(
    # crystal
    lamb0=None,
    bragg0=None,
    # geometry
    xx=None,
    length=None,
    rcurve=None,
    dist=None,
    beta_max=None,
    # computed
    ap=None,
    crystx=None,
    crysty=None,
    endx=None,
    endy=None,
    ilamb_min=None,
    ilamb_max=None,
    lamb_min=None,
    lamb_max=None,
    Dlamb=None,
    x0=None,
    # camera
    dcam=None,
    # plotting
    dax=None,
    # saving
    pfe_fig=None,
    # unused
    **kwdargs,
):

    # ----------
    # prepare

    npts, size = crystx.shape

    # envelop
    iok = np.isfinite(endx)
    i0 = tuple([iok[:, ii].nonzero()[0][0] for ii in range(size)])
    i1 = tuple([iok[:, ii].nonzero()[0][-1] for ii in range(size)])
    nind = tuple(range(size))

    # envelop
    envx = np.array([
        endx[i1, nind], crystx[i1, nind],
        np.full((size,), ap[0]),
        crystx[i0, nind], endx[i0, nind],
    ])
    envy = np.array([
        endy[i1, nind], crysty[i1, nind],
        np.full((size,), ap[1]),
        crysty[i0, nind], endy[i0, nind],
    ])

    # central rays
    ind = int((npts-1)/2)
    raycx = np.array([np.full((size,), ap[0]), crystx[ind, :], endx[ind, :]])
    raycy = np.array([np.full((size,), ap[1]), crysty[ind, :], endy[ind, :]])

    # dcam
    if dcam is not None:
        ninx, niny = dcam['nin_r'][:2]
        ninn = np.sqrt(ninx**2 + niny**2)
        ninx, niny = ninx/ninn, niny/ninn
        e0x, e0y = -niny, ninx
        e0n = np.sqrt(e0x**2 + e0y**2)
        e0x, e0y = e0x/e0n, e0y/e0n
        clen = dcam['length']
        camx = dcam['cent_r'][0] + 0.5*clen*np.r_[-1, 1] * e0x
        camy = dcam['cent_r'][1] + 0.5*clen*np.r_[-1, 1] * e0y

    # --------------
    # prepare figure

    if dax is None:
        dax = _get_axes()

    # -----------
    # plot

    color = None
    for ii in range(size):

        # ---
        # hor

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            # crystals
            ll, = ax.plot(
                crystx[:, ii],
                crysty[:, ii],
                ls='-',
                lw=2,
                marker='None',
            )
            color = ll.get_color()

            # central rays
            ax.plot(
                raycx[:, ii],
                raycy[:, ii],
                ls='--',
                lw=1,
                marker='None',
                c=color,
            )

            # edge rays
            ax.plot(
                envx[:, ii],
                envy[:, ii],
                ls='-',
                lw=1,
                marker='None',
                c=color,
                label=(
                    f"r = {rcurve[ii]} m\t"
                    + r"$\lambda_0$" + f" = {lamb0[ii]*1e10:5.3f} AA\t"
                    + r"$\beta_0$" + f" = {bragg0[ii]*180/np.pi:5.2f} deg"
                ),
            )

        kax = 'cam'
        if dcam is not None and dax.get(kax) is not None:
            ax = dax[kax]['handle']

            # images
            ax.plot(
                dcam['x0'][:, ii],
                np.full((npts,), ii+1),
                ls='None',
                marker='.',
                color=color,
                ms=6,
            )

            # lamb min, max
            ax.text(
                dcam['x0'][ilamb_min[ii], ii],
                ii + 1 - 0.1,
                f'{lamb_min[ii]*1e10:2.3} AA',
                color=color,
                size=8,
                horizontalalignment='center',
                verticalalignment='top',
            )

            ax.text(
                dcam['x0'][ilamb_max[ii], ii],
                ii + 1 - 0.1,
                f'{lamb_max[ii]*1e10:2.3} AA',
                color=color,
                size=8,
                horizontalalignment='center',
                verticalalignment='top',
            )

    # ---------------
    # plot input data

    kax = 'hor'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']
        ax.legend(fontsize=12)

    if beta_max is None:
        beta_str = 'None'
    else:
        beta_str = f'{beta_max*180/np.pi:5.3} deg'

    msg = (
        f"beta_max = {beta_str}\n"
    )

    ax.text(
        0.8,
        0.4,
        msg,
        color='k',
        size=10,
        horizontalalignment='center',
        verticalalignment='top',
        transform=ax.figure.transFigure,
    )


    # ------------
    # camera

    if dcam is not None:

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            ax.plot(
                camx,
                camy,
                ls='-',
                lw=2.,
                marker='None',
                c='k',
            )

        kax = 'cam'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            ax.axvline(-0.5*dcam['length'], c='k', ls='-', lw=1.)
            ax.axvline(0.5*dcam['length'], c='k', ls='-', lw=1.)
            ax.set_ylim(0, size + 1)

    # ----------
    # saving

    if pfe_fig is not None:
        dax['hor']['handle'].figure.savefig(pfe_fig, format='png', dpi=200)

    return dax


def _get_axes():

    # --------------
    # prepare figure

    dmargin = {
        'left': 0.08, 'right': 0.98,
        'bottom': 0.08, 'top': 0.90,
        'hspace': 0.20, 'wspace': 0.25,
    }

    fig = plt.figure(figsize=(13, 8))
    fig.suptitle('2d ray-tracing model')
    gs = gridspec.GridSpec(ncols=3, nrows=2, **dmargin)

    # ----------
    # make axes

    # ax0 - hor
    ax0 = fig.add_subplot(
        gs[:, :-1],
        aspect='equal',
        adjustable='datalim',
    )

    ax0.set_xlabel("x (m)", size=12)
    ax0.set_ylabel("y (m)", size=12)
    ax0.set_title("2d ray tracing", size=12, fontweight='bold')

    # ax1 - cam
    ax1 = fig.add_subplot(
        gs[0, -1],
        aspect='auto',
    )

    ax1.set_xlabel("x0 (m)", size=12)
    ax1.set_title("Image on camera", size=12, fontweight='bold')

    # ------------
    # dict

    dax = {
        'hor': {'handle': ax0},
        'cam': {'handle': ax1},
    }

    return dax


# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import os


import numpy as np
import matplotlib.pyplot as plt


_PATH_HERE = os.path.dirname(__file__)
_PATH_TOFU = os.path.dirname(os.path.dirname(_PATH_HERE))

import tofu as tf


_PATH_TESTDATA_01 = os.path.join(
    os.path.dirname(_PATH_HERE),
    'tests',
    'tests01_geom',
    'test_data',
)
_PATH_TESTDATA_04 = os.path.join(
    os.path.dirname(_PATH_HERE),
    'tests',
    'tests04_spectro',
    'test_data',
)
_PFE_TESTDATA = os.path.join(_PATH_TESTDATA_04, 'spectral_fit.npz')
_PFE_DET = os.path.join(_PATH_TESTDATA_01, 'det37_CTVD_incC4_New.npz')
_PFE_CRYST = os.path.join(
    _PATH_TESTDATA_04,
    'TFG_CrystalBragg_ExpWEST_DgXICS_ArXVII_sh00000_Vers1.5.0.npz',
)


# #############################################################################
# #############################################################################
#               Benchmark of Mesh2D routines
#                   High level routines
# #############################################################################


class HighLevel:
    """ Benchmark the geometry-oriented routines

    In particular:
        - camera creation (ray-tracing)
        - solid angle computing

    """

    # -----------------------------
    # Attributes reckognized by asv

    # time before benchmark is killed
    timeout = 500
    repeat = (1, 10, 20.0)
    sample_time = 0.100

    # -------------------------------------------------------
    # Setup and teardown, run before / after benchmark methods

    def setup_cache(self):
        return 0

    def setup(self, out):
        """ run before each benchmark method, out from setup_cache  """
        out = dict(np.load(_PFE_TESTDATA, allow_pickle=True))
        self.lamb = out['lamb']
        self.var = out['var']
        self.dlines = out['dlines'].tolist()
        self.spect2d = out['spect2d']

        self.conf0 = tf.load_config('WEST-V0')
        self.cryst = tf.load(_PFE_CRYST)
        self.det = dict(np.load(_PFE_DET, allow_pickle=True))
        self.xixj_lim = [
            [-0.041882, 0.041882], 0.1 + 100*172.e-6*np.r_[-0.5, 0.5]
        ]

    def teardown(self, out):
        pass

    # -------------------------------------
    # benchmarks methods for geometry tools

    # Mesh2D - rect
    def time_00_fit1d(self, out):

        # Define constraint dict
        dconst = {
            'amp': {'a1': ['a', 'd']},
            'width': 'group',
            'shift': {
                'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                'b': {'key': 's1', 'coef': 1., 'offset': 0.},
                'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                'd': {'key': 's3', 'coef': 1., 'offset': 0.001e-10},
            },
            'double': True,
            'symmetry': False,
        }

        dx0 = {
            'width': 1.,
            'shift': {
                's1': 0.,
                's2': 1.,
            },
            'dratio': 0,
            'dshift': 0,
        }

        domain = {
            'lamb': [
                [3.94e-10, 3.952e-10],
                (3.95e-10, 3.956e-10),
                [3.96e-10, 4e-10],
            ],
        }

        data = self.spect2d[:5, :]
        pos = True
        focus = 'a'
        dconstants = {'shift': {'s1': 0}}

        dinput = tf.spectro.fit1d_dinput(
            dlines=self.dlines,
            dconstraints=dconst,
            dconstants=dconstants,
            dprepare=None,
            data=data,
            lamb=self.lamb,
            mask=None,
            domain=domain,
            pos=pos,
            subset=None,
            same_spectrum=None,
            nspect=None,
            same_spectrum_dlamb=None,
            focus=focus,
            valid_fraction=0.28,     # fraction of pixels ok per time step
            valid_nsigma=0,         # S/N ratio for each pixel
            focus_half_width=None,
            valid_return_fract=None,
            dscales=None,
            dx0=dx0,
            dbounds=None,
            # defconst=None,
        )

        dfit1d = tf.spectro.fit1d(
            dinput=dinput,
            method=None,
            Ti=None,
            chain=True,
            jac='dense',
            verbose=False,
            plot=False,
        )

    def time_01_fit2d(self, out):
        # Define constraint dict
        dconst = {
            'amp': {'a1': ['a', 'd']},
            'width': 'group',
            'shift': {
                'a': {'key': 's1', 'coef': 1., 'offset': 0.},
                'b': {'key': 's1', 'coef': 1., 'offset': 0.},
                'c': {'key': 's2', 'coef': 2., 'offset': 0.},
                'd': {'key': 's3', 'coef': 1., 'offset': 0.001e-10},
            },
            'double': False,
            'symmetry': False,
        }

        dx0 = {
            'width': 1.,
            'shift': {
                's1': 0.,
                's2': 1.,
            },
            'dratio': 0,
            'dshift': 0,
        }

        domain = {
            'lamb': [
                [3.94e-10, 3.952e-10],
                (3.95e-10, 3.956e-10),
                [3.96e-10, 4e-10],
            ],
        }

        data = self.spect2d
        pos = True
        focus = 'a'
        dconstants = {'shift': {'s1': 0}}

        dinput = tf.spectro.fit2d_dinput(
            dlines=self.dlines,
            dconstraints=dconst,
            dconstants=dconstants,
            dprepare=None,
            data=data,
            lamb=self.lamb,
            phi=np.arange(0, data.shape[0]),
            binning=None,
            mask=None,
            domain=domain,
            pos=pos,
            focus=focus,
            valid_fraction=0.28,     # fraction of pixels ok per time step
            valid_nsigma=0,         # S/N ratio for each pixel
            focus_half_width=None,
            valid_return_fract=None,
            dscales=None,
            dx0=dx0,
            dbounds=None,
            # defconst=None,
        )

        dfit2d = tf.spectro.fit2d(
            dinput=dinput,
            method=None,
            Ti=None,
            chain=True,
            jac='dense',
            xtol=1e-6,
            ftol=1e-6,
            gtol=1e-6,
            verbose=False,
            plot=False,
        )


    def time_02_get_plasmadomain_at_lamb(self, out):
        pts, lambok = self.cryst.get_plasmadomain_at_lamb(
            det=self.det,
            lamb=[3.94e-10, 4.e-10],
            res=[0.005, 0.005, 0.01],
            config=self.conf0,
            domain=[None, [-0.36, -0.22], [-4*np.pi/5., -np.pi/2.]],
            xixj_lim=self.xixj_lim,
            plot=False,
        )
